import pytest
import numpy as np
from .._acquisition_functions import ACQUISITION_FUNCTIONS


@pytest.mark.parametrize("name", ACQUISITION_FUNCTIONS.keys())
def test_acquisition_functions(name):
    func = ACQUISITION_FUNCTIONS[name]
    probabs = np.array([[0.5, 0.5], [0.01, 0.99]])
    uncertainty = func(probabs)
    if name == "random":
        assert (uncertainty == 0.5).all()
    else:
        assert uncertainty[0] > uncertainty[1]



"""
Download, transform and simulate various binary datasets.
"""

# Author: Georgios Douzas <gdouzas@icloud.com>
#         Joao Fonseca <jpfonseca@novaims.unl.pt>
# License: MIT

from re import sub
from collections import Counter
from itertools import product
from urllib.parse import urljoin
from string import ascii_lowercase
from zipfile import ZipFile
from io import BytesIO, StringIO

import requests
import numpy as np
import pandas as pd
from sklearn.utils import check_X_y
from imblearn.datasets import make_imbalance

from .base import Datasets, FETCH_URLS, RANDOM_STATE


class ImbalancedBinaryDatasets(Datasets):
    """Class to download, transform and save binary class imbalanced
    datasets."""

    MULTIPLICATION_FACTORS = [2, 3]

    @staticmethod
    def _calculate_ratio(multiplication_factor, y):
        """Calculate ratio based on IRs multiplication factor."""
        ratio = Counter(y).copy()
        ratio[1] = int(ratio[1] / multiplication_factor)
        return ratio

    def _make_imbalance(self, data, multiplication_factor):
        """Undersample the minority class."""
        X_columns = [col for col in data.columns if col != "target"]
        X, y = check_X_y(data.loc[:, X_columns], data.target)
        if multiplication_factor > 1.0:
            sampling_strategy = self._calculate_ratio(multiplication_factor, y)
            X, y = make_imbalance(
                X, y, sampling_strategy=sampling_strategy, random_state=RANDOM_STATE
            )
        data = pd.DataFrame(np.column_stack((X, y)))
        data[data.columns[-1]] = data[data.columns[-1]].astype(int)
        return data

    def download(self):
        """Download the datasets and append undersampled versions of them."""
        super(ImbalancedBinaryDatasets, self).download()
        undersampled_datasets = []
        for (name, data), factor in list(
            product(self.content_, self.MULTIPLICATION_FACTORS)
        ):
            ratio = self._calculate_ratio(factor, data.target)
            if ratio[1] >= 15:
                data = self._make_imbalance(data, factor)
                undersampled_datasets.append((f"{name} ({factor})", data))
        self.content_ += undersampled_datasets
        return self

    def fetch_breast_tissue(self):
        """Download and transform the Breast Tissue Data Set.
        The minority class is identified as the `car` and `fad`
        labels and the majority class as the rest of the labels.

        http://archive.ics.uci.edu/ml/datasets/breast+tissue
        """
        data = pd.read_excel(FETCH_URLS["breast_tissue"], sheet_name="Data")
        data = data.drop(columns="Case #").rename(columns={"Class": "target"})
        data["target"] = data["target"].isin(["car", "fad"]).astype(int)
        return data

    def fetch_ecoli(self):
        """Download and transform the Ecoli Data Set.
        The minority class is identified as the `pp` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/ecoli
        """
        data = pd.read_csv(FETCH_URLS["ecoli"], header=None, delim_whitespace=True)
        data = data.drop(columns=0).rename(columns={8: "target"})
        data["target"] = data["target"].isin(["pp"]).astype(int)
        return data

    def fetch_eucalyptus(self):
        """Download and transform the Eucalyptus Data Set.
        The minority class is identified as the `best` label
        and the majority class as the rest of the labels.

        https://www.openml.org/d/188
        """
        data = pd.read_csv(FETCH_URLS["eucalyptus"])
        data = data.iloc[:, -9:].rename(columns={"Utility": "target"})
        data = data[data != "?"].dropna()
        data["target"] = data["target"].isin(["best"]).astype(int)
        return data

    def fetch_glass(self):
        """Download and transform the Glass Identification Data Set.
        The minority class is identified as the `1` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/glass+identification
        """
        data = pd.read_csv(FETCH_URLS["glass"], header=None)
        data = data.drop(columns=0).rename(columns={10: "target"})
        data["target"] = data["target"].isin([1]).astype(int)
        return data

    def fetch_haberman(self):
        """Download and transform the Haberman's Survival Data Set.
        The minority class is identified as the `1` label
        and the majority class as the `0` label.

        https://archive.ics.uci.edu/ml/datasets/Haberman's+Survival
        """
        data = pd.read_csv(FETCH_URLS["haberman"], header=None)
        data.rename(columns={3: "target"}, inplace=True)
        data["target"] = data["target"].isin([2]).astype(int)
        return data

    def fetch_heart(self):
        """Download and transform the Heart Data Set.
        The minority class is identified as the `2` label
        and the majority class as the `1` label.

        http://archive.ics.uci.edu/ml/datasets/statlog+(heart)
        """
        data = pd.read_csv(FETCH_URLS["heart"], header=None, delim_whitespace=True)
        data.rename(columns={13: "target"}, inplace=True)
        data["target"] = data["target"].isin([2]).astype(int)
        return data

    def fetch_iris(self):
        """Download and transform the Iris Data Set.
        The minority class is identified as the `1` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/iris
        """
        data = pd.read_csv(FETCH_URLS["iris"], header=None)
        data.rename(columns={4: "target"}, inplace=True)
        data["target"] = data["target"].isin(["Iris-setosa"]).astype(int)
        return data

    def fetch_libras(self):
        """Download and transform the Libras Movement Data Set.
        The minority class is identified as the `1` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/Libras+Movement
        """
        data = pd.read_csv(FETCH_URLS["libras"], header=None)
        data.rename(columns={90: "target"}, inplace=True)
        data["target"] = data["target"].isin([1]).astype(int)
        return data

    def fetch_liver(self):
        """Download and transform the Liver Disorders Data Set.
        The minority class is identified as the `1` label
        and the majority class as the '2' label.

        https://archive.ics.uci.edu/ml/datasets/liver+disorders
        """
        data = pd.read_csv(FETCH_URLS["liver"], header=None)
        data.rename(columns={6: "target"}, inplace=True)
        data["target"] = data["target"].isin([1]).astype(int)
        return data

    def fetch_pima(self):
        """Download and transform the Pima Indians Diabetes Data Set.
        The minority class is identified as the `1` label
        and the majority class as the '0' label.

        https://www.kaggle.com/uciml/pima-indians-diabetes-database
        """
        data = pd.read_csv(FETCH_URLS["pima"], header=None, skiprows=9)
        data.rename(columns={8: "target"}, inplace=True)
        return data

    def fetch_vehicle(self):
        """Download and transform the Vehicle Silhouettes Data Set.
        The minority class is identified as the `1` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/Statlog+(Vehicle+Silhouettes)
        """
        data = []
        for letter in ascii_lowercase[0:9]:
            partial_data = pd.read_csv(
                urljoin(FETCH_URLS["vehicle"], "xa%s.dat" % letter),
                header=None,
                delim_whitespace=True,
            )
            partial_data = partial_data.rename(columns={18: "target"})
            partial_data["target"] = partial_data["target"].isin(["van"]).astype(int)
            data.append(partial_data)
        return pd.concat(data)

    def fetch_wine(self):
        """Download and transform the Wine Data Set.
        The minority class is identified as the `2` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/wine
        """
        data = pd.read_csv(FETCH_URLS["wine"], header=None)
        data.rename(columns={0: "target"}, inplace=True)
        data["target"] = data["target"].isin([2]).astype(int)
        return data

    def fetch_new_thyroid_1(self):
        """Download and transform the Thyroid Disease Data Set.
        The minority class is identified as the `positive`
        label and the majority class as the `negative` label.

        .. note:: The positive class was originally label 2.

        https://archive.ics.uci.edu/ml/datasets/Thyroid+Disease
        """
        data = pd.read_csv(
            FETCH_URLS["new_thyroid"],
            header=None,
        )
        data.rename(columns={0: "target"}, inplace=True)
        data["target"] = (data["target"] == 2).astype(int)
        return data

    def fetch_new_thyroid_2(self):
        """Download and transform the Thyroid Disease Data Set.
        The minority class is identified as the `positive`
        label and the majority class as the `negative` label.

        .. note:: The positive class was originally label 3.

        https://archive.ics.uci.edu/ml/datasets/Thyroid+Disease
        """
        data = pd.read_csv(
            FETCH_URLS["new_thyroid"],
            header=None,
        )
        data.rename(columns={0: "target"}, inplace=True)
        data["target"] = (data["target"] == 3).astype(int)
        return data

    def fetch_cleveland(self):
        """Download and transform the Heart Disease Cleveland Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        https://archive.ics.uci.edu/ml/datasets/heart+disease
        """
        data = pd.read_csv(FETCH_URLS["cleveland"], header=None, na_values="?")
        data.dropna(inplace=True)
        data.rename(columns={13: "target"}, inplace=True)
        data["target"] = (data["target"] == 1).astype(int)
        return data

    def fetch_dermatology(self):
        """Download and transform the Dermatology Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        https://archive.ics.uci.edu/ml/datasets/Dermatology
        """
        data = pd.read_csv(FETCH_URLS["dermatology"], header=None)
        data.rename(columns={34: "target"}, inplace=True)
        data.drop(columns=33, inplace=True)
        data["target"] = (data.target == 6).astype(int)
        return data

    def fetch_led(self):
        """Download and transform the LED Display Domain Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        https://www.openml.org/d/40496
        """
        data = pd.read_csv(FETCH_URLS["led"])
        data.rename(columns={"Class": "target"}, inplace=True)
        data["target"] = (data.target == 1).astype(int)
        return data

    def fetch_page_blocks(self):
        """Download and transform the Page Blocks Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        https://www.openml.org/d/1021
        """
        data = pd.read_csv(FETCH_URLS["page_blocks"])
        data.rename(columns={"class": "target"}, inplace=True)
        data["target"] = (data.target != 1).astype(int)
        return data

    def fetch_vowel(self):
        """Download and transform the Vowel Recognition Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        https://www.openml.org/d/375
        """

        data = pd.read_csv(FETCH_URLS["vowels"])
        data.rename(columns={"speaker": "target"}, inplace=True)
        data.drop(columns=["utterance", "frame"], inplace=True)
        data["target"] = (data["target"] == 1).astype(int)
        return data

    def fetch_yeast(self):
        """Download and transform the Yeast Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        https://archive.ics.uci.edu/ml/datasets/Yeast
        """
        data = pd.read_csv(FETCH_URLS["yeast"], header=None)
        data = pd.DataFrame(
            [
                [val for val in row.split(" ") if len(val) != 0]
                for row in data[0].tolist()
            ]
        )
        data.drop(columns=0, inplace=True)
        data.rename(columns={9: "target"}, inplace=True)
        data["target"] = (data["target"] == "MIT").astype(int)
        return data


class BinaryDatasets(Datasets):
    """Class to download, transform and save binary class datasets."""

    def fetch_banknote_authentication(self):
        """Download and transform the Banknote Authentication Data Set.

        https://archive.ics.uci.edu/ml/datasets/banknote+authentication
        """
        data = pd.read_csv(FETCH_URLS["banknote_authentication"], header=None)
        data.rename(columns={4: "target"}, inplace=True)
        return data

    def fetch_arcene(self):
        """Download and transform the Arcene Data Set.

        https://archive.ics.uci.edu/ml/datasets/Arcene
        """
        url = FETCH_URLS["arcene"]
        data, labels = [], []
        for data_type in ("train", "valid"):
            data.append(
                pd.read_csv(
                    urljoin(url, f"ARCENE/arcene_{data_type}.data"),
                    header=None,
                    sep=" ",
                ).drop(columns=list(range(1998, 10001)))
            )
            labels.append(
                pd.read_csv(
                    urljoin(
                        url,
                        ("ARCENE/" if data_type == "train" else "")
                        + f"arcene_{data_type}.labels",
                    ),
                    header=None,
                ).rename(columns={0: "target"})
            )
        data = pd.concat(data, ignore_index=True)
        labels = pd.concat(labels, ignore_index=True)
        data = pd.concat([data, labels], axis=1)
        data["target"] = data["target"].isin([1]).astype(int)
        return data

    def fetch_audit(self):
        """Download and transform the Audit Data Set.

        https://archive.ics.uci.edu/ml/datasets/Audit+Data
        """
        zipped_data = requests.get(FETCH_URLS["audit"]).content
        unzipped_data = (
            ZipFile(BytesIO(zipped_data))
            .read("audit_data/audit_risk.csv")
            .decode("utf-8")
        )
        data = pd.read_csv(StringIO(sub(r"@.+\n+", "", unzipped_data)), engine="python")
        data = (
            data.drop(columns=["LOCATION_ID"])
            .rename(columns={"Risk": "target"})
            .dropna()
        )
        return data

    def fetch_spambase(self):
        """Download and transform the Spambase Data Set.

        https://archive.ics.uci.edu/ml/datasets/Spambase
        """
        data = pd.read_csv(FETCH_URLS["spambase"], header=None)
        data.rename(columns={57: "target"}, inplace=True)
        return data

    def fetch_parkinsons(self):
        """Download and transform the Parkinsons Data Set.

        https://archive.ics.uci.edu/ml/datasets/parkinsons
        """
        data = pd.read_csv(FETCH_URLS["parkinsons"])
        data = pd.concat(
            [
                data.drop(columns=["name", "status"]),
                data[["status"]].rename(columns={"status": "target"}),
            ],
            axis=1,
        )
        data["target"] = data["target"].isin([0]).astype(int)
        return data

    def fetch_ionosphere(self):
        """Download and transform the Ionosphere Data Set.

        https://archive.ics.uci.edu/ml/datasets/ionosphere
        """
        data = pd.read_csv(FETCH_URLS["ionosphere"], header=None)
        data = data.drop(columns=[0, 1]).rename(columns={34: "target"})
        data["target"] = data["target"].isin(["b"]).astype(int)
        return data

    def fetch_breast_cancer(self):
        """Download and transform the Breast Cancer Wisconsin Data Set.

        https://archive.ics.uci.edu/ml/datasets/Breast+Cancer+Wisconsin+(Diagnostic)
        """
        data = pd.read_csv(FETCH_URLS["breast_cancer"], header=None)
        data = pd.concat(
            [data.drop(columns=[0, 1]), data[[1]].rename(columns={1: "target"})], axis=1
        )
        data["target"] = data["target"].isin(["M"]).astype(int)
        return data



"""Testing active learning models."""
import pytest
from itertools import product
import re
import numpy as np
from sklearn import datasets
from sklearn.utils.validation import check_random_state
from sklearn.utils._testing import assert_array_equal, ignore_warnings
from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import RandomForestClassifier

from ...data_augmentation import OverSamplingAugmentation, GeometricSMOTE
from ...metrics import get_scorer
from .._acquisition_functions import ACQUISITION_FUNCTIONS
from .._active_learning import StandardAL, AugmentationAL

ACTIVE_LEARNERS = {"StandardAL": StandardAL, "AugmentationAL": AugmentationAL}

RANDOM_STATE = 42

# toy sample
X = [[-2, -1], [-1, -1], [-1, -2], [1, 1], [1, 2], [2, 1]]
y = [-1, -1, -1, 1, 1, 1]
T = [[-1, -1], [2, 2], [3, 2]]
true_result = [-1, 1, 1]

# Larger classification sample used for testing feature importances
X_large, y_large = datasets.make_classification(
    n_samples=100,
    n_features=10,
    n_informative=3,
    n_redundant=0,
    n_repeated=0,
    shuffle=False,
    random_state=RANDOM_STATE,
)

# also load the iris dataset
# and randomly permute it
iris = datasets.load_iris()
rng = check_random_state(RANDOM_STATE)
perm = rng.permutation(iris.target.size)
iris.data = iris.data[perm]
iris.target = iris.target[perm]

# also make a hastie_10_2 dataset
hastie_X, hastie_y = datasets.make_hastie_10_2(n_samples=20, random_state=RANDOM_STATE)
hastie_X = hastie_X.astype(np.float32)

DATASETS = [(X, y), (iris.data, iris.target), (hastie_X, hastie_y)]


@pytest.mark.parametrize("name", ACTIVE_LEARNERS.keys())
def test_classification_toy(name):
    """Test classification on a toy dataset."""
    al_model = ACTIVE_LEARNERS[name](random_state=RANDOM_STATE)
    al_model.fit(X, y)
    assert_array_equal(al_model.predict(T), true_result)
    assert al_model.random_state == RANDOM_STATE
    assert al_model.classifier_.random_state == RANDOM_STATE


@pytest.mark.parametrize("name", ACTIVE_LEARNERS.keys())
@pytest.mark.parametrize("X, y", DATASETS)
def test_default_parameters(name, X, y):
    """Test default parameters."""
    al_model = ACTIVE_LEARNERS[name]()
    al_model.fit(X, y)

    # Set up expected parameters
    exp_acquisition_func = ACQUISITION_FUNCTIONS["random"]
    exp_n_init = int(np.round(0.02 * np.array(X).shape[0]))
    exp_n_init = exp_n_init if exp_n_init >= 2 else 2
    exp_budget = int(np.round(0.02 * np.array(X).shape[0]))
    exp_budget = exp_budget if exp_budget >= 1 else 1
    exp_max_iter = int((np.array(X).shape[0] - exp_n_init) / exp_budget)

    assert type(al_model.classifier_) == RandomForestClassifier
    assert al_model.acquisition_func_ == exp_acquisition_func
    assert al_model.n_init_ == exp_n_init
    assert al_model.budget_ == exp_budget
    assert al_model.max_iter_ == exp_max_iter
    assert al_model.random_state is None
    assert al_model.continue_training is False


def test_augmentation_active_learning():
    """Test active learning with pipelined data augmentation and parameter tuning."""
    generator = OverSamplingAugmentation(GeometricSMOTE(n_jobs=-1))
    classifier = RandomForestClassifier(n_jobs=-1)
    al_model = AugmentationAL(
        classifier=classifier,
        generator=generator,
        param_grid={"generator__value": [1, 1.5, 2]},
        max_iter=2,
        random_state=RANDOM_STATE,
    )
    al_model.fit(iris.data, iris.target)

    assert al_model.max_iter_ == 2
    assert al_model.random_state == RANDOM_STATE
    assert (
        dict(al_model.classifier_.best_estimator_.steps)["generator"].random_state
        == RANDOM_STATE
    )


@pytest.mark.parametrize("name", ACTIVE_LEARNERS.keys())
def test_classifier_metadata(name):
    al_model = ACTIVE_LEARNERS[name](random_state=RANDOM_STATE)
    al_model.fit(X, y, X_test=T, y_test=true_result)

    metadata = al_model.metadata_
    exp_metadata_keys = ["data", "performance_metric", *range(5)]

    assert al_model._has_test
    assert list(metadata.keys()) == exp_metadata_keys


@pytest.mark.parametrize("name", ACTIVE_LEARNERS.keys())
@ignore_warnings
def test_al_params(name):
    """Test less commonly used parameters"""
    classifier = MLPClassifier(max_iter=2)

    # passing a function instead of a string and set continued training
    al_model = ACTIVE_LEARNERS[name](
        classifier=classifier,
        acquisition_func=ACQUISITION_FUNCTIONS["entropy"],
        continue_training=True,
        max_iter=5,
        n_init=2,
        budget=1,
    )
    al_model.fit(X, y)

    # passing various types of values for initialization and budget
    for param, n in product(["n_init", "budget"], [0.01, 2]):
        al_model = ACTIVE_LEARNERS[name](
            classifier=classifier,
            acquisition_func=ACQUISITION_FUNCTIONS["entropy"],
            n_init=2,
            budget=1,
            max_iter=5,
        )
        al_model.set_params(**{param: n})
        al_model.fit(X, y)

        assert al_model.get_params()[param] == n
        assert getattr(al_model, "n_init_") == 2
        if param == "n_init" or al_model.budget == 0.01:
            assert getattr(al_model, "budget_") == 1
        else:
            assert getattr(al_model, "budget_") == 2

    # passing an evaluation metric as string
    ACTIVE_LEARNERS[name](
        classifier=classifier,
        evaluation_metric="f1_macro",
        n_init=2,
        budget=1,
        max_iter=5,
    ).fit(X, y)

    # passing an evaluation metric as function
    ACTIVE_LEARNERS[name](
        classifier=classifier,
        evaluation_metric=get_scorer("f1_macro"),
        n_init=2,
        budget=1,
        max_iter=5,
    ).fit(X, y)

    # fit with test set
    ACTIVE_LEARNERS[name](
        classifier=classifier, n_init=2, budget=1, max_iter=5, continue_training=False
    ).fit(X, y, X_test=T, y_test=true_result).predict(T)


@pytest.mark.parametrize("name", ACTIVE_LEARNERS.keys())
@ignore_warnings
def test_errors(name):
    classifier = MLPClassifier(max_iter=2)
    err_msg = "``continue_training`` must be of type ``bool``. Got err instead."
    with pytest.raises(TypeError, match=re.escape(err_msg)):
        ACTIVE_LEARNERS[name](
            classifier=classifier, max_iter=5, continue_training="err"
        ).fit(X, y)

    al_model = ACTIVE_LEARNERS[name](classifier=classifier, max_iter=5)
    err_msg = f"Active Learning model {type(al_model).__name__} is not initialized yet."
    with pytest.raises(StopIteration, match=re.escape(err_msg)):
        al_model.iteration(X, y)

    err_msg = f"Active Learning model {type(al_model).__name__} is already initialized."
    with pytest.raises(StopIteration, match=re.escape(err_msg)):
        al_model.fit(X, y).fit(X, y)



"""Toolbox to develop research in Machine Learning.

``research`` is a library containing the implementation of various algorithms developed
in Machine Learning research, as well as utilities to facilitate the formatting of pandas
dataframes into LaTeX tables.

Subpackages
-----------
active_learning
    Module which contains the code developed for experiments related to Active Learning.
data_augmentation
    Module which contains the implementation of variations of oversampling/data
    augmentation algorithms, as well as helper classes to use oversampling algorithms as
    data augmentation techniques.
datasets
    Module which contains code to download, transform and simulate various datasets.
metrics
    Module which contains performance metrics/scorers that are not
    included in scikit-learn's scorers' dictionary.
preprocessing
    Contains preprocessing methods adapted or modified from sklearn.
latex
    This module contains several functions to prepare and format tables for LaTeX
    documents.
utils
    Contains a variety of general utility functions and tools used to format and prepare
    tables to incorporate into LaTeX code.
"""
import sys

try:
    # This variable is injected in the __builtins__ by the build
    # process. It is used to enable importing subpackages of sklearn when
    # the binaries are not built
    # mypy error: Cannot determine type of '__SKLEARN_SETUP__'
    __MLRESEARCH_SETUP__  # type: ignore
except NameError:
    __MLRESEARCH_SETUP__ = False

if __MLRESEARCH_SETUP__:
    sys.stderr.write("Partial import of imblearn during the build process.\n")
    # We are not importing the rest of scikit-learn during the build
    # process, as it may not be compiled yet
else:
    from . import active_learning
    from . import data_augmentation
    from . import datasets
    from . import metrics
    from . import preprocessing
    from . import utils

    from ._version import __version__

    __all__ = [
        "active_learning",
        "data_augmentation",
        "datasets",
        "metrics",
        "preprocessing",
        "utils",
        "__version__",
    ]



"""
Utility functions related to image processing
"""

import numpy as np
import pandas as pd


def image_to_dataframe(X, y=None, bands=None, target_feature="target"):
    """
    Converts an image array (height, width, bands) to a pandas dataframe
    (height * width, bands). If ``y`` is not ``None``, a feature with name
    ``target_feature`` will be added to the dataset.

    .. note:: Some workflows use image arrays of format (bands, width, height). In that
        case, you can simply transpose the image before calling this function.

    Parameters
    ----------
    X : array-like of shape (h, w, b)
        Matrix containing the image data.

    y : array-like of shape (h, w), default=None
        The target values (class labels) as integers or strings.

    bands : array-like of shape (b,), default=None
        The names of the bands in the image.

    target_feature : str, default="target"
        Target feature name.

    Returns
    -------
    df_image : pd.DataFrame, shape (h * w, b[+1])
        Dataframe with pixel coordinates (h, w) as index, counting from the top left
        corner.

    Examples
    --------

    >>> import numpy as np
    >>> X = np.random.default_rng(42).random((4,5,3))
    >>> y = np.random.default_rng(42).integers(0,2,4*5).reshape(4,5)
    >>> image_to_dataframe(X).head(5)
    ...             0         1         2
    ... h w
    ... 0 0  0.773956  0.438878  0.858598
    ...   1  0.697368  0.094177  0.975622
    ...   2  0.761140  0.786064  0.128114
    ...   3  0.450386  0.370798  0.926765
    ...   4  0.643865  0.822762  0.443414
    >>> image_to_dataframe(X, y, bands=["r","g", "b"], target_feature="classes").head(5)
    ...             r         g         b  classes
    ... h w
    ... 0 0  0.773956  0.438878  0.858598        0
    ...   1  0.697368  0.094177  0.975622        1
    ...   2  0.761140  0.786064  0.128114        1
    ...   3  0.450386  0.370798  0.926765        0
    ...   4  0.643865  0.822762  0.443414        0
    """
    # Check y's dimensionality
    if y is not None and len(y.shape) == 2:
        y = np.expand_dims(y, axis=-1)

    # Collect metadata
    shp = X.shape
    columns = [i for i in range(shp[-1])] if bands is None else bands
    indices = np.indices(shp[:-1]).reshape((2, shp[0] * shp[1]))
    indices = pd.MultiIndex.from_arrays(indices, names=["h", "w"])

    if y is None:
        dat = np.moveaxis(X, -1, 0).reshape((len(columns), shp[0] * shp[1]))
        df_image = pd.DataFrame(data=dat.T, columns=columns, index=indices)
    else:
        columns = columns + [target_feature]
        dat = np.moveaxis(np.append(X, y, axis=-1), -1, 0).reshape(
            (len(columns), shp[0] * shp[1])
        )
        df_image = pd.DataFrame(data=dat.T, columns=columns, index=indices)
        df_image[target_feature] = df_image[target_feature].astype(int)

    return df_image


def dataframe_to_image(df, bands=None, target_feature=None):
    """
    Converts a pandas dataframe to an image. The height ("h"), and width ("w")
    coordinates of the image must be in the index.

    Arguments
    ---------
    df : pd.DataFrame
        Dataframe with pixel coordinates (h, w) as index, counting from the top left
        corner.

    bands : array-like of shape (b,), default=None
        The names of the bands in the dataframe to be passed to the image.

    target_feature : str, default=None
        Target feature name.

    Returns
    -------
    X : array-like of shape (h, w, b)
        Matrix containing the image data.

    y : array-like of shape (h, w), default=None
        The target values (class labels) as integers or strings.

    bands : array-like of shape (b,), default=None
        The names of the bands in the image.
    """

    if (np.sort(df.index.names) != ["h", "w"]).any():
        err_msg = (
            'Image coordinates with names ["h", "w"] must be in index,'
            f" got {list(df.index.names)} instead."
        )
        raise IndexError(err_msg)

    df_ = df.reset_index().pivot(index="h", columns="w")
    y = df_[target_feature].values if target_feature is not None else None

    if bands is None:
        bands = (
            df.columns if target_feature is None else df.columns.drop(target_feature)
        )

    X = np.array([df_[band] for band in bands])
    X = np.moveaxis(X, 0, -1)

    return X, y, bands



"""
Class to perform over-sampling using Geometric SMOTE. This is a modified
version of the original Geometric SMOTE implementation.
"""

# Author: Georgios Douzas <gdouzas@icloud.com>
#         Joao Fonseca    <jpmrfonseca@gmail.com>
# License: BSD 3 clause

import math
import numpy as np
from collections import Counter
from numpy.linalg import norm
from scipy import sparse
from sklearn.utils import check_random_state, check_array
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import _check_sample_weight
from sklearn.utils.sparsefuncs_fast import (
    csr_mean_variance_axis0,
    csc_mean_variance_axis0,
)
from sklearn.preprocessing import OneHotEncoder, label_binarize
from imblearn.over_sampling.base import BaseOverSampler
from imblearn.utils import (
    check_neighbors_object,
    Substitution,
    check_target_type,
    check_sampling_strategy,
)
from imblearn.utils._docstring import _random_state_docstring
from imblearn.utils._validation import ArraysTransformer

SELECTION_STRATEGY = ("combined", "majority", "minority")


def _make_geometric_sample(
    center, surface_point, truncation_factor, deformation_factor, random_state
):
    """A support function that returns an artificial point inside
    the geometric region defined by the center and surface points.

    Parameters
    ----------
    center : ndarray, shape (n_features, )
        Center point of the geometric region.

    surface_point : ndarray, shape (n_features, )
        Surface point of the geometric region.

    truncation_factor : float, optional (default=0.0)
        The type of truncation. The values should be in the [-1.0, 1.0] range.

    deformation_factor : float, optional (default=0.0)
        The type of geometry. The values should be in the [0.0, 1.0] range.

    random_state : int, RandomState instance or None
        Control the randomization of the algorithm.

    Returns
    -------
    point : ndarray, shape (n_features, )
            Synthetically generated sample.

    """

    # Zero radius case
    if np.array_equal(center, surface_point):
        return center

    # Generate a point on the surface of a unit hyper-sphere
    radius = norm(center - surface_point)
    normal_samples = random_state.normal(size=center.size)
    point_on_unit_sphere = normal_samples / norm(normal_samples)
    point = (random_state.uniform(size=1) ** (1 / center.size)) * point_on_unit_sphere

    # Parallel unit vector
    parallel_unit_vector = (surface_point - center) / norm(surface_point - center)

    # Truncation
    close_to_opposite_boundary = (
        truncation_factor > 0
        and np.dot(point, parallel_unit_vector) < truncation_factor - 1
    )
    close_to_boundary = (
        truncation_factor < 0
        and np.dot(point, parallel_unit_vector) > truncation_factor + 1
    )
    if close_to_opposite_boundary or close_to_boundary:
        point -= 2 * np.dot(point, parallel_unit_vector) * parallel_unit_vector

    # Deformation
    parallel_point_position = np.dot(point, parallel_unit_vector) * parallel_unit_vector
    perpendicular_point_position = point - parallel_point_position
    point = (
        parallel_point_position
        + (1 - deformation_factor) * perpendicular_point_position
    )

    # Translation
    point = center + radius * point

    return point


def _make_categorical_sample(X_new, all_neighbors, categories_size, random_state):
    """A support function that populates categorical features' values
    in an artificial point.

    Parameters
    ----------
    X_new : ndarray, shape (n_features, )
        Artificial point to populate categorical features.

    all_neighbors: ndarray, shape (n_features, k_neighbors)
        Nearest neighbors used for majority voting.

    categories_size: list
        Used to tell apart one-hot encoded features.

    random_state : int, RandomState instance or None
        Control the randomization of the algorithm. Used
        for tie breaking when there are two majority values.

    Returns
    -------
    point : ndarray, shape (n_features, )
            Synthetically generated sample.

    """
    for start_idx, end_idx in zip(
        np.cumsum(categories_size)[:-1], np.cumsum(categories_size)[1:]
    ):
        col_maxs = all_neighbors[:, start_idx:end_idx].sum(axis=0)
        # tie breaking argmax
        is_max = np.isclose(col_maxs, col_maxs.max(axis=0))
        max_idxs = random_state.permutation(np.argwhere(is_max))
        col_sels = max_idxs[0]

        ys = start_idx + col_sels
        X_new[start_idx:end_idx] = 0
        X_new[ys] = 1

    return X_new


@Substitution(
    sampling_strategy=BaseOverSampler._sampling_strategy_docstring,
    random_state=_random_state_docstring,
)
class GeometricSMOTE(BaseOverSampler):
    """Class to to perform over-sampling using Geometric SMOTE.

    This algorithm is an implementation of Geometric SMOTE, a geometrically
    enhanced drop-in replacement for SMOTE as presented in [1]_.

    Parameters
    ----------
    categorical_features : ndarray of shape (n_cat_features,) or (n_features,)
        Specified which features are categorical. Can either be:

        - array of indices specifying the categorical features;
        - mask array of shape (n_features, ) and ``bool`` dtype for which
          ``True`` indicates the categorical features.

    {sampling_strategy}

    {random_state}

    truncation_factor : float, optional (default=0.0)
        The type of truncation. The values should be in the [-1.0, 1.0] range.

    deformation_factor : float, optional (default=0.0)
        The type of geometry. The values should be in the [0.0, 1.0] range.

    selection_strategy : str, optional (default='combined')
        The type of Geometric SMOTE algorithm with the following options:
        ``'combined'``, ``'majority'``, ``'minority'``.

    k_neighbors : int or object, optional (default=5)
        If ``int``, number of nearest neighbours to use when synthetic
        samples are constructed for the minority method.  If object, an estimator
        that inherits from :class:`sklearn.neighbors.base.KNeighborsMixin` that
        will be used to find the k_neighbors.

    n_jobs : int, optional (default=1)
        The number of threads to open if possible.

    Attributes
    ----------
    sampling_strategy_ : dict
        Dictionary containing the information to sample the dataset. The keys
        corresponds to the class labels from which to sample and the values
        are the number of samples to sample.
    n_features_in_ : int
        Number of features in the input dataset.
    nns_pos_ : estimator object
        Validated k-nearest neighbours created from the `k_neighbors` parameter. It is
        used to find the nearest neighbors of the same class of a selected
        observation.
    nn_neg_ : estimator object
        Validated k-nearest neighbours created from the `k_neighbors` parameter. It is
        used to find the nearest neighbor of the remaining classes (k=1) of a selected
        observation.
    random_state_ : instance of RandomState
        If the `random_state` parameter is None, it is a RandomState singleton used by
        np.random. If `random_state` is an int, it is a RandomState instance seeded with
        seed. If `random_state` is already a RandomState instance, it is the same
        object.

    Notes
    -----
    See the original paper: [1]_ for more details.

    Supports multi-class resampling. A one-vs.-rest scheme is used as
    originally proposed in [2]_.

    References
    ----------

    .. [1] G. Douzas, F. Bacao, "Geometric SMOTE:
       a geometrically enhanced drop-in replacement for SMOTE",
       Information Sciences, vol. 501, pp. 118-135, 2019.

    .. [2] N. V. Chawla, K. W. Bowyer, L. O. Hall, W. P. Kegelmeyer, "SMOTE:
       synthetic minority over-sampling technique", Journal of Artificial
       Intelligence Research, vol. 16, pp. 321-357, 2002.

    Examples
    --------

    >>> from collections import Counter
    >>> from sklearn.datasets import make_classification
    >>> from gsmote import GeometricSMOTE # doctest: +NORMALIZE_WHITESPACE
    >>> X, y = make_classification(n_classes=2, class_sep=2,
    ... weights=[0.1, 0.9], n_informative=3, n_redundant=1, flip_y=0,
    ... n_features=20, n_clusters_per_class=1, n_samples=1000, random_state=10)
    >>> print('Original dataset shape %s' % Counter(y))
    Original dataset shape Counter({{1: 900, 0: 100}})
    >>> gsmote = GeometricSMOTE(random_state=1)
    >>> X_res, y_res = gsmote.fit_resample(X, y)
    >>> print('Resampled dataset shape %s' % Counter(y_res))
    Resampled dataset shape Counter({{0: 900, 1: 900}})

    """

    def __init__(
        self,
        sampling_strategy="auto",
        random_state=None,
        truncation_factor=1.0,
        deformation_factor=0.0,
        selection_strategy="combined",
        k_neighbors=5,
        categorical_features=None,
        n_jobs=1,
    ):
        super(GeometricSMOTE, self).__init__(sampling_strategy=sampling_strategy)
        self.random_state = random_state
        self.truncation_factor = truncation_factor
        self.deformation_factor = deformation_factor
        self.selection_strategy = selection_strategy
        self.k_neighbors = k_neighbors
        self.categorical_features = categorical_features
        self.n_jobs = n_jobs

    def _validate_estimator(self):
        """Create the necessary attributes for Geometric SMOTE."""

        # Check random state
        self.random_state_ = check_random_state(self.random_state)

        # Validate strategy
        if self.selection_strategy not in SELECTION_STRATEGY:
            error_msg = (
                "Unknown selection_strategy for Geometric SMOTE algorithm. "
                "Choices are {}. Got {} instead."
            )
            raise ValueError(
                error_msg.format(SELECTION_STRATEGY, self.selection_strategy)
            )

        # Create nearest neighbors object for positive class
        if self.selection_strategy in ("minority", "combined"):
            self.nns_pos_ = check_neighbors_object(
                "nns_positive", self.k_neighbors, additional_neighbor=1
            )
            self.nns_pos_.set_params(n_jobs=self.n_jobs)

        # Create nearest neighbors object for negative class
        if self.selection_strategy in ("majority", "combined"):
            self.nn_neg_ = check_neighbors_object("nn_negative", nn_object=1)
            self.nn_neg_.set_params(n_jobs=self.n_jobs)

    def _validate_categorical(self):
        """Create the necessary attributes for Geometric SMOTE
        with categorical features"""

        if self.categorical_features is None:
            return self

        categorical_features = np.asarray(self.categorical_features)
        if categorical_features.dtype.name == "bool":
            self.categorical_features_ = np.flatnonzero(categorical_features)
        else:
            if any(
                [
                    cat not in np.arange(self.n_features_in_)
                    for cat in categorical_features
                ]
            ):
                raise ValueError(
                    "Some of the categorical indices are out of range. Indices"
                    " should be between 0 and {}".format(self.n_features_in_)
                )
            self.categorical_features_ = categorical_features
        self.continuous_features_ = np.setdiff1d(
            np.arange(self.n_features_in_), self.categorical_features_
        )

        if self.categorical_features_.size == self.n_features_in_:
            raise ValueError(
                "GeometricSMOTE is not designed to work only with categorical "
                "features. It requires some numerical features."
            )
        return self

    def _check_X_y(self, X, y):
        """Overwrite the checking to let pass some string for categorical
        features.
        """
        y, binarize_y = check_target_type(y, indicate_one_vs_all=True)
        X, y = self._validate_data(
            X, y, reset=True, dtype=None, accept_sparse=["csr", "csc"]
        )

        return X, y, binarize_y

    def _make_geometric_samples(
        self, X, y, pos_class_label, n_samples, sample_weight=None
    ):
        """A support function that returns an artificials samples inside
        the geometric region defined by nearest neighbors.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Matrix containing the data which have to be sampled.
        y : array-like, shape (n_samples, )
            Corresponding label for each sample in X.
        pos_class_label : str or int
            The minority class (positive class) target value.
        n_samples : int
            The number of samples to generate.
        sample_weight : 1-D array-like, optional
            The probabilities associated with each entry in a.
            If not given, the sample assumes a uniform distribution over all
            entries.

        Returns
        -------
        X_new : ndarray, shape (n_samples_new, n_features)
            Synthetically generated samples.
        y_new : ndarray, shape (n_samples_new, )
            Target values for synthetic samples.

        """

        # Return zero new samples
        if n_samples == 0:
            return (
                np.array([], dtype=X.dtype).reshape(0, X.shape[1]),
                np.array([], dtype=y.dtype),
                np.array([], dtype=X.dtype),
            )

        # Select positive class samples
        X_pos = X[y == pos_class_label]
        if sample_weight is not None:
            sample_weight_pos = (
                sample_weight[y == pos_class_label]
                if sample_weight[y == pos_class_label].sum() != 0
                else None
            )
        else:
            sample_weight_pos = None

        # Force minority strategy if no negative class samples are present
        self.selection_strategy_ = (
            "minority" if X.shape[0] == X_pos.shape[0] else self.selection_strategy
        )

        # Minority or combined strategy
        if self.selection_strategy_ in ("minority", "combined"):
            self.nns_pos_.fit(X_pos)
            points_pos = self.nns_pos_.kneighbors(X_pos)[1][:, 1:]
            weight_pos = (
                np.repeat(sample_weight_pos, self.k_neighbors)
                / (sample_weight_pos.sum() * self.k_neighbors)
                if sample_weight_pos is not None
                else None
            )
            samples_indices = self.random_state_.choice(
                range(0, len(points_pos.flatten())), size=n_samples, p=weight_pos
            )
            rows = np.floor_divide(samples_indices, points_pos.shape[1])
            cols = np.mod(samples_indices, points_pos.shape[1])

        # Majority or combined strategy
        if self.selection_strategy_ in ("majority", "combined"):
            X_neg = X[y != pos_class_label]
            self.nn_neg_.fit(X_neg)
            points_neg = self.nn_neg_.kneighbors(X_pos)[1]
            weight_neg = (
                sample_weight_pos / sample_weight_pos.sum()
                if sample_weight_pos is not None
                else None
            )
            if self.selection_strategy_ == "majority":
                samples_indices = self.random_state_.choice(
                    range(0, len(points_neg.flatten())), size=n_samples, p=weight_neg
                )
                rows = np.floor_divide(samples_indices, points_neg.shape[1])
                cols = np.mod(samples_indices, points_neg.shape[1])

        # In the case that the median std was equal to zeros, we have to
        # create non-null entry based on the encoded of OHE
        if self.categorical_features is not None:
            if math.isclose(self.median_std_, 0):
                X[:, self.continuous_features_.size :] = self._X_categorical_encoded
                # Select positive class samples
                X_pos = X[y == pos_class_label]
                if self.selection_strategy_ in ("majority", "combined"):
                    X_neg = X[y != pos_class_label]

        # Generate new samples
        X_new = np.zeros((n_samples, X.shape[1]))
        all_neighbors_ = []
        for ind, (row, col) in enumerate(zip(rows, cols)):

            # Define center point
            center = X_pos[row]

            # Minority strategy
            if self.selection_strategy_ == "minority":
                surface_point = X_pos[points_pos[row, col]]
                all_neighbors = (
                    (X_pos[points_pos[row]])
                    if self.categorical_features is not None
                    else None
                )

            # Majority strategy
            elif self.selection_strategy_ == "majority":
                surface_point = X_neg[points_neg[row, col]]
                all_neighbors = (
                    (X_neg[points_neg[row]])
                    if self.categorical_features is not None
                    else None
                )

            # Combined strategy
            else:
                surface_point_pos = X_pos[points_pos[row, col]]
                surface_point_neg = X_neg[points_neg[row, 0]]
                radius_pos = norm(center - surface_point_pos)
                radius_neg = norm(center - surface_point_neg)
                surface_point = (
                    surface_point_neg if radius_pos > radius_neg else surface_point_pos
                )
                all_neighbors = (
                    np.vstack([X_pos[points_pos[row]], X_neg[points_neg[row]]])
                    if self.categorical_features is not None
                    else None
                )

            if self.categorical_features is not None:
                all_neighbors_.append(all_neighbors)

            # Append new sample - no categorical features
            X_new[ind] = _make_geometric_sample(
                center,
                surface_point,
                self.truncation_factor,
                self.deformation_factor,
                self.random_state_,
            )

        # Create new samples for target variable
        y_new = np.array([pos_class_label] * len(samples_indices))

        return X_new, y_new, all_neighbors_

    def _make_categorical_samples(self, X_new, y_new, categories_size, all_neighbors_):
        for ind, all_neighbors in enumerate(all_neighbors_):
            # Append new sample - continuous features
            X_new[ind] = _make_categorical_sample(
                X_new[ind], all_neighbors, categories_size, self.random_state_
            )
        return X_new, y_new

    def _encode_categorical(self, X, y):
        """
        One-Hot encodes categorical features and replaces the 1 entries with the median
        of the standard deviations divided by 2.
        """
        # compute the median of the standard deviation of the minority class
        target_stats = Counter(y)
        class_minority = min(target_stats, key=target_stats.get)

        # Separate categorical features from continuous features
        X_continuous = X[:, self.continuous_features_]
        X_continuous = check_array(X_continuous, accept_sparse=["csr", "csc"])
        X_categorical = X[:, self.categorical_features_].copy()
        X_minority = X_continuous[np.flatnonzero(y == class_minority)]

        if sparse.issparse(X):
            if X.format == "csr":
                _, var = csr_mean_variance_axis0(X_minority)
            else:
                _, var = csc_mean_variance_axis0(X_minority)
        else:
            var = X_minority.var(axis=0)
        self.median_std_ = np.median(np.sqrt(var))

        if X_continuous.dtype.name != "object":
            dtype_ohe = X_continuous.dtype
        else:
            dtype_ohe = np.float64
        self.ohe_ = OneHotEncoder(
            sparse_output=True, handle_unknown="ignore", dtype=dtype_ohe
        )

        # the input of the OneHotEncoder needs to be dense
        X_ohe = self.ohe_.fit_transform(
            X_categorical.toarray() if sparse.issparse(X_categorical) else X_categorical
        )

        # we can replace the 1 entries of the categorical features with the
        # median of the standard deviation. It will ensure that whenever
        # distance is computed between 2 samples, the difference will be equal
        # to the median of the standard deviation as in the original paper.

        # In the edge case where the median of the std is equal to 0, the 1s
        # entries will be also nullified. In this case, we store the original
        # categorical encoding which will be later used for inversing the OHE
        if math.isclose(self.median_std_, 0):
            self._X_categorical_encoded = X_ohe.toarray()

        X_ohe.data = np.ones_like(X_ohe.data, dtype=X_ohe.dtype) * self.median_std_ / 2

        if self._issparse:
            X_encoded = np.hstack([X_continuous.toarray(), X_ohe.toarray()])
        else:
            X_encoded = np.hstack([X_continuous, X_ohe.toarray()])

        return X_encoded

    def _decode_categorical(self, X_resampled):
        """Reverses the encoding of the categorical features to match
        the dataset's original structure."""

        if math.isclose(self.median_std_, 0):
            X_resampled[
                : self._X_categorical_encoded.shape[0], self.continuous_features_.size :
            ] = self._X_categorical_encoded

        X_resampled = sparse.csr_matrix(X_resampled)

        X_res_cat = X_resampled[:, self.continuous_features_.size :]
        X_res_cat.data = np.ones_like(X_res_cat.data)
        X_res_cat_dec = self.ohe_.inverse_transform(X_res_cat)

        if self._issparse:
            X_resampled = sparse.hstack(
                (X_resampled[:, : self.continuous_features_.size], X_res_cat_dec),
                format="csr",
            )
        else:
            X_resampled = np.hstack(
                (
                    X_resampled[:, : self.continuous_features_.size].toarray(),
                    X_res_cat_dec,
                )
            )

        indices_reordered = np.argsort(
            np.hstack((self.continuous_features_, self.categorical_features_))
        )

        if sparse.issparse(X_resampled):
            col_indices = X_resampled.indices.copy()
            for idx, col_idx in enumerate(indices_reordered):
                mask = X_resampled.indices == col_idx
                col_indices[mask] = idx
            X_resampled.indices = col_indices
        else:
            X_resampled = X_resampled[:, indices_reordered]

        return X_resampled

    def _fit_resample(self, X, y, sample_weight=None):

        # Save basic data
        self._issparse = sparse.issparse(X)
        X_dtype = X.dtype

        # Validate estimator's parameters
        self._validate_categorical()._validate_estimator()

        # Preprocess categorical data
        if self.categorical_features is not None:
            X = self._encode_categorical(X, y)
            categories_size = [self.continuous_features_.size] + [
                cat.size for cat in self.ohe_.categories_
            ]

        # Copy data
        X_resampled, y_resampled = X.copy(), y.copy()

        # Resample
        for class_label, n_samples in self.sampling_strategy_.items():

            # Apply gsmote mechanism
            X_new, y_new, all_neighbors_ = self._make_geometric_samples(
                X, y, class_label, n_samples, sample_weight=sample_weight
            )

            # Apply smotenc mechanism
            if self.categorical_features is not None:
                X_new, y_new = self._make_categorical_samples(
                    X_new, y_new, categories_size, all_neighbors_
                )

            # Append new data
            X_resampled, y_resampled = (
                np.vstack((X_resampled, X_new)),
                np.hstack((y_resampled, y_new)),
            )

        # reverse the encoding of the categorical features
        if self.categorical_features is not None:
            X_resampled = self._decode_categorical(X_resampled).astype(X_dtype)
        else:
            X_resampled = X_resampled.astype(X_dtype)
        return X_resampled, y_resampled

    def fit_resample(self, X, y, sample_weight=None):
        """Resample the dataset.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Matrix containing the data which have to be sampled.

        y : array-like of shape (n_samples,)
            Corresponding label for each sample in X.

        sample_weight : array-like of shape (n_samples,), default=None
            Individual weights for each sample. Assigns probabilities for selecting a
            sample as a center point.

        Returns
        -------
        X_resampled : {array-like, sparse matrix} of shape \
                (n_samples_new, n_features)
            The array containing the resampled data.

        y_resampled : array-like of shape (n_samples_new,)
            The corresponding label of `X_resampled`.
        """
        check_classification_targets(y)
        arrays_transformer = ArraysTransformer(X, y)
        X, y, binarize_y = self._check_X_y(X, y)

        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X, dtype=X.dtype)

        self.sampling_strategy_ = check_sampling_strategy(
            self.sampling_strategy, y, self._sampling_type
        )

        output = self._fit_resample(X, y, sample_weight)

        y_ = label_binarize(output[1], np.unique(y)) if binarize_y else output[1]

        X_, y_ = arrays_transformer.transform(output[0], y_)
        return (X_, y_) if len(output) == 2 else (X_, y_, output[2])



"""Test performance metrics"""

# Author: Joao Fonseca <jpfonseca@novaims.unl.pt>

import pytest
import numpy as np
from sklearn.utils._testing import ignore_warnings

from ...active_learning import StandardAL
from .._metrics import (
    ALScorer,
    geometric_mean_score_macro,
    area_under_learning_curve,
    data_utilization_rate,
)

RANDOM_STATE = 42

# toy sample
X = [[-2, -1], [-1, -1], [-1, -2], [1, 1], [1, 2], [2, 1]]
y = [-1, -1, -1, 1, 1, 1]
T = [[-1, -1], [2, 2], [3, 2]]
true_result = [-1, 1, 1]


def test_geometric_mean_score_macro_binary():
    # Dense label indicator matrix format
    y1 = np.array([0, 1, 1])
    y2 = np.array([0, 0, 1])

    assert geometric_mean_score_macro(y1, y2) == 0.75
    assert geometric_mean_score_macro(y1, y1) == 1
    assert geometric_mean_score_macro(y2, y2) == 1
    assert geometric_mean_score_macro(y2, np.logical_not(y2)) == 0
    assert geometric_mean_score_macro(y1, np.logical_not(y1)) == 0
    assert geometric_mean_score_macro(y1, np.zeros(y1.shape)) == 0.5
    assert geometric_mean_score_macro(y2, np.zeros(y1.shape)) == 0.5


@ignore_warnings
def test_geometric_mean_score_macro_single_class():
    # Test G-mean behavior with a single positive or negative class
    # Such a case may occur with non-stratified cross-validation
    assert 0.0 == geometric_mean_score_macro([1, 1], [1, 1])
    assert 0.0 == geometric_mean_score_macro([-1, -1], [-1, -1])


def test_geometric_mean_score_macro_multiclass():
    # Dense label indicator matrix format
    y1 = np.array([0, 1, 2, 0, 1, 2])
    y2 = np.array([0, 2, 1, 0, 0, 1])

    assert geometric_mean_score_macro(y1, y2) == pytest.approx(0.471, rel=1e-2)


def test_al_metrics():
    al_model = StandardAL(random_state=RANDOM_STATE)
    al_model.fit(X, y, X_test=T, y_test=true_result)

    assert ALScorer(data_utilization_rate)(al_model, None, None) == 1 / 3
    assert ALScorer(area_under_learning_curve)(al_model, None, None) == 1



"""
Download, transform and simulate various datasets.
"""
from ._binary import (
    ImbalancedBinaryDatasets,
    BinaryDatasets,
)
from ._multiclass import (
    ContinuousCategoricalDatasets,
    MultiClassDatasets,
)
from ._remote_sensing import RemoteSensingDatasets

__all__ = [
    "ImbalancedBinaryDatasets",
    "BinaryDatasets",
    "ContinuousCategoricalDatasets",
    "MultiClassDatasets",
    "RemoteSensingDatasets",
]



from itertools import product
from sklearn.base import clone
from sklearn.utils import check_random_state
from sklearn.preprocessing import FunctionTransformer
from sklearn.model_selection import ParameterGrid
from imblearn.utils import Substitution
from imblearn.pipeline import Pipeline
from imblearn.utils._docstring import _random_state_docstring


@Substitution(
    random_state=_random_state_docstring,
)
def check_random_states(random_state, n_runs):
    """
    Create random states for experiments. Used to create seeds for different
    initializations.

    Arguments
    ---------
    {random_state}

    n_runs : int
        Number of initializations.

    Returns
    -------
    random_states : list
        A list of random states with length ``n_runs``.
    """
    random_state = check_random_state(random_state)
    return [random_state.randint(0, 2**32 - 1, dtype="uint32") for _ in range(n_runs)]


@Substitution(
    random_state=_random_state_docstring,
)
def check_pipelines(*objects_list, random_state, n_runs):
    """
    Extract estimators and parameter grids to be passed to ModelSearchCV. This enables
    searching over any sequence of parameter settings and objects.

    Arguments
    ---------
    *objects_list : sequence of lists
        Lists of objects to be chained in a pipeline in the passed order. Each list
        must contain tuples composed of (``<obj_name>``, ``<object>``,
        ``<parameter_values_dict>``).

    {random_state}

    n_runs : int
        Number of initializations.

    Returns
    -------
    estimators : List of Pipelines with all combinations among the passed lists of
        objects.

    param_grids : List of dictionaries with estimator and parameter names (``str``) as
        keys and lists of parameter settings to try as values.
    """

    # Create random states
    random_states = check_random_states(random_state, n_runs)

    pipelines = []
    param_grid = []
    for comb, rs in product(product(*objects_list), random_states):
        name = "|".join([i[0] for i in comb])

        # name, object, sub grid
        comb = [
            (nm, ob, ParameterGrid(sg))
            if ob is not None
            else (nm, FunctionTransformer(), ParameterGrid(sg))
            for nm, ob, sg in comb
        ]

        # Create estimator
        if name not in [n[0] for n in pipelines]:
            est = Pipeline([(nm, ob) for nm, ob, _ in comb])
            pipelines.append((name, est))

        # Create intermediate parameter grids
        sub_grids = [
            [{f"{nm}__{k}": v for k, v in param_def.items()} for param_def in sg]
            for nm, obj, sg in comb
        ]

        # Create parameter grids
        for sub_grid in product(*sub_grids):
            param_prefix = f"{name}__"
            grid = {"est_name": [name]}
            grid.update(
                {f"{param_prefix}{k}": [v] for d in sub_grid for k, v in d.items()}
            )
            random_states = {
                f"{param_prefix}{param}": [rs]
                for param in est.get_params()
                if "random_state" in param
            }
            grid.update(random_states)

            # Avoid multiple runs over pipelines without random state
            if grid not in param_grid:
                param_grid.append(grid)

    return pipelines, param_grid


def check_pipelines_wrapper(
    *objects_list,
    wrapper,
    random_state,
    n_runs,
    estimator_param="classifier",
    wrapped_only=True,
):
    """
    Extract estimators within a wrapper object and parameter grids to be passed to
    ModelSearchCV. This enables searching over any sequence of parameter settings and
    objects.

    Arguments
    ---------
    *objects_list : sequence of lists
        Lists of objects to be chained in a pipeline in the passed order. Each list
        must contain tuples composed of (``<obj_name>``, ``<object>``,
        ``<parameter_values_dict>``).

    wrapper : tuple or tuple
        Wrapper object to which the lists of objects will be passed. Must be structured
        as (``<obj_name>``, ``<object>``, ``<parameter_values_dict>``) and .

    {random_state}

    n_runs : int
        Number of initializations.

    estimator_param : str, default="classifier"
        Name of the parameter in the wrapper object where the estimators will be passed.

    wrapped_only : bool, default=True
        Return only the wrapped estimators. If ``False``, returns both the wrapped and
        the original objects.

    Returns
    -------
    wrapped_estimators : List of Pipelines with all combinations among the passed lists
        of objects.

    wrapped_param_grids : List of dictionaries with estimator and parameter names
        (``str``) as keys and lists of parameter settings to try as values.
    """

    wrapper_label = wrapper[0]
    wrapper_obj = wrapper[1]
    wrapper_grid = wrapper[2]

    estimators, param_grids = check_pipelines(
        *objects_list, random_state=random_state, n_runs=n_runs
    )

    wrapped_estimators = [
        (
            f"{wrapper_label}|{name}",
            clone(wrapper_obj).set_params(**{estimator_param: pipeline}),
        )
        for name, pipeline in estimators
    ]

    def _format_param(param):
        return "__".join(param.split("__")[1:])

    wrapped_param_grids = [
        {
            "est_name": [f'{wrapper_label}|{d["est_name"][0]}'],
            **{
                f'{wrapper_label}|{d["est_name"][0]}__{estimator_param}__{_format_param(k)}': v
                for k, v in d.items()
                if k != "est_name"
            },
            **{
                f'{wrapper_label}|{d["est_name"][0]}__{k}': v
                for k, v in wrapper_grid.items()
            },
        }
        for d in param_grids
    ]
    if wrapped_only:
        return wrapped_estimators, wrapped_param_grids
    else:
        return (estimators + wrapped_estimators, param_grids + wrapped_param_grids)



"""All minimum dependencies for ml-research."""
import argparse

NUMPY_MIN_VERSION = "1.14.6"
PANDAS_MIN_VERSION = "1.3.5"
SKLEARN_MIN_VERSION = "1.0.0"
IMBLEARN_MIN_VERSION = "0.8.0"
RICH_MIN_VERSION = "10.16.1"
MATPLOTLIB_MIN_VERSION = "2.2.3"

# The values are (version_spec, comma separated tags)
dependent_packages = {
    "pandas": (PANDAS_MIN_VERSION, "install"),
    "numpy": (NUMPY_MIN_VERSION, "install"),
    "scikit-learn": (SKLEARN_MIN_VERSION, "install"),
    "imbalanced-learn": (IMBLEARN_MIN_VERSION, "install"),
    "rich": (RICH_MIN_VERSION, "install"),
    "requests": ("2.26.0", "install"),
    "matplotlib": (MATPLOTLIB_MIN_VERSION, "optional, tests"),
    "research-learn": ("0.3.0", "tests"),
    "pytest-cov": ("3.0.0", "tests"),
    "flake8": ("3.8.2", "tests"),
    "black": ("22.3", "tests"),
    "pylint": ("2.12.2", "tests"),
    "coverage": ("6.2", "tests"),
    "sphinx": ("4.2.0", "docs"),
    "numpydoc": ("1.0.0", "docs, tests"),
    "sphinx-material": ("0.0.35", "docs"),
    "nbsphinx": ("0.8.7", "docs"),
    "recommonmark": ("0.7.1", "docs"),
    "sphinx-markdown-tables": ("0.0.15", "docs"),
    "sphinx-copybutton": ("0.4.0", "docs"),
}


# create inverse mapping for setuptools
tag_to_packages: dict = {
    extra: [] for extra in ["install", "optional", "docs", "examples", "tests", "all"]
}
for package, (min_version, extras) in dependent_packages.items():
    for extra in extras.split(", "):
        tag_to_packages[extra].append("{}>={}".format(package, min_version))
    tag_to_packages["all"].append("{}>={}".format(package, min_version))


# Used by CI to get the min dependencies
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Get min dependencies for a package")

    parser.add_argument("package", choices=dependent_packages)
    args = parser.parse_args()
    min_version = dependent_packages[args.package][0]
    print(min_version)



"""
Test the check_pipelines module.
"""
from itertools import product

from sklearn.tree import DecisionTreeClassifier
from sklearn.preprocessing import MinMaxScaler
from sklearn.decomposition import PCA
from sklearn.datasets import load_iris
from imblearn.over_sampling import SMOTE, BorderlineSMOTE
from imblearn.pipeline import Pipeline
from rlearn.model_selection import ModelSearchCV

from .._check_pipelines import (
    check_pipelines,
    check_pipelines_wrapper,
    check_random_states,
)
from ...active_learning import StandardAL
from ...data_augmentation import OverSamplingAugmentation


def test_check_pipeline_single():
    """Test the check of pipelines with a single element."""

    # Initialization
    n_runs = 5
    rnd_seed = 0
    classifiers = [("clf", DecisionTreeClassifier(), {"max_depth": [3, 5]})]

    # Estimators and parameters grids
    estimators, param_grids = check_pipelines(
        classifiers, random_state=rnd_seed, n_runs=n_runs
    )
    names, pips = zip(*estimators)
    steps = [
        [(step[0], step[1].__class__.__name__) for step in pip.steps] for pip in pips
    ]

    # Expected estimators and parameters grids
    exp_name = "clf"
    exp_steps = [("clf", "DecisionTreeClassifier")]
    exp_random_states = check_random_states(rnd_seed, n_runs)
    partial_param_grids = []
    for max_depth in [3, 5]:
        partial_param_grids.append(
            {
                "clf__clf__max_depth": [max_depth],
            }
        )

    exp_param_grids = []
    for rnd_seed, partial_param_grid in product(exp_random_states, partial_param_grids):
        partial_param_grid = partial_param_grid.copy()
        partial_param_grid.update(
            {
                "est_name": ["clf"],
                "clf__clf__random_state": [rnd_seed],
            }
        )
        exp_param_grids.append(partial_param_grid)

    # Assertions
    assert names[0] == exp_name
    assert steps[0] == exp_steps
    assert len(param_grids) == len(exp_param_grids)
    assert all([param_grid in exp_param_grids for param_grid in param_grids])


def test_check_pipelines():
    """Test the check of oversamplers and classifiers."""

    # Initialization
    n_runs = 5
    rnd_seed = 0
    oversamplers = [
        ("ovs", BorderlineSMOTE(), [{"k_neighbors": [2, 4]}, {"m_neighbors": [6, 8]}])
    ]
    classifiers = [("clf", DecisionTreeClassifier(), {"max_depth": [3, 5]})]

    # Estimators and parameters grids
    estimators, param_grids = check_pipelines(
        oversamplers, classifiers, random_state=rnd_seed, n_runs=n_runs
    )
    names, pips = zip(*estimators)
    steps = [
        [(step[0], step[1].__class__.__name__) for step in pip.steps] for pip in pips
    ]

    # Expected estimators and parameters grids
    exp_name = "ovs|clf"
    exp_steps = [("ovs", "BorderlineSMOTE"), ("clf", "DecisionTreeClassifier")]
    exp_random_states = check_random_states(rnd_seed, n_runs)
    partial_param_grids = []
    for k_neighbors, max_depth in product([2, 4], [3, 5]):
        partial_param_grids.append(
            {
                "ovs|clf__ovs__k_neighbors": [k_neighbors],
                "ovs|clf__clf__max_depth": [max_depth],
            }
        )
    for m_neighbors, max_depth in product([6, 8], [3, 5]):
        partial_param_grids.append(
            {
                "ovs|clf__ovs__m_neighbors": [m_neighbors],
                "ovs|clf__clf__max_depth": [max_depth],
            }
        )
    exp_param_grids = []
    for rnd_seed, partial_param_grid in product(exp_random_states, partial_param_grids):
        partial_param_grid = partial_param_grid.copy()
        partial_param_grid.update(
            {
                "est_name": ["ovs|clf"],
                "ovs|clf__ovs__random_state": [rnd_seed],
                "ovs|clf__clf__random_state": [rnd_seed],
            }
        )
        exp_param_grids.append(partial_param_grid)

    # Assertions
    assert names[0] == exp_name
    assert steps[0] == exp_steps
    assert len(param_grids) == len(exp_param_grids)
    assert all([param_grid in exp_param_grids for param_grid in param_grids])


def test_check_oversamplers_classifiers_none():
    """Test the check of oversamplers and classifiers for no oversampler."""

    # Initialization
    n_runs = 2
    rnd_seed = 12
    oversamplers = [("ovs", None, {})]
    classifiers = [("clf", DecisionTreeClassifier(), {"max_depth": [3, 5, 8]})]

    # Estimators and parameters grids
    estimators, param_grids = check_pipelines(
        oversamplers, classifiers, random_state=rnd_seed, n_runs=n_runs
    )
    names, pips = zip(*estimators)
    steps = [
        [(step[0], step[1].__class__.__name__) for step in pip.steps] for pip in pips
    ]

    # Expected names, steps and parameter grids
    exp_name = "ovs|clf"
    exp_steps = [("ovs", "FunctionTransformer"), ("clf", "DecisionTreeClassifier")]
    exp_random_states = check_random_states(rnd_seed, n_runs)
    partial_param_grids = []
    for max_depth in [3, 5, 8]:
        partial_param_grids.append({"ovs|clf__clf__max_depth": [max_depth]})
    exp_param_grids = []
    for rnd_seed, partial_param_grid in product(exp_random_states, partial_param_grids):
        partial_param_grid = partial_param_grid.copy()
        partial_param_grid.update(
            {"est_name": ["ovs|clf"], "ovs|clf__clf__random_state": [rnd_seed]}
        )
        exp_param_grids.append(partial_param_grid)

    # Assertions
    assert names[0] == exp_name
    assert steps[0] == exp_steps
    assert len(param_grids) == len(exp_param_grids)
    assert all([param_grid in exp_param_grids for param_grid in param_grids])


def test_check_oversamplers_classifiers_pipeline():
    """Test the check of pipelines with higher lengths."""

    # Initialization
    n_runs = 2
    rnd_seed = 3
    scalers = [("scaler", MinMaxScaler(), {"feature_range": [(0, 1), (0, 0.5)]})]
    oversamplers = [("ovs", SMOTE(), {"k_neighbors": [3, 5]})]
    classifiers = [
        (
            "clf",
            Pipeline([("pca", PCA()), ("dtc", DecisionTreeClassifier())]),
            {"pca__n_components": [4, 8], "dtc__max_depth": [3, 5]},
        )
    ]

    # Estimators and parameters grids
    estimators, param_grids = check_pipelines(
        scalers, oversamplers, classifiers, random_state=rnd_seed, n_runs=n_runs
    )
    names, pips = zip(*estimators)
    steps = [
        [(step[0], step[1].__class__.__name__) for step in pip.steps] for pip in pips
    ]

    # Expected names, steps and parameter grids
    exp_name = "scaler|ovs|clf"
    exp_steps = [
        ("scaler", "MinMaxScaler"),
        ("ovs", "SMOTE"),
        ("clf", "Pipeline"),
    ]
    exp_random_states = check_random_states(rnd_seed, n_runs)
    partial_param_grids = []
    for feature_range, k_neighbors, n_components, max_depth in product(
        [(0, 1), (0, 0.5)], [3, 5], [4, 8], [3, 5]
    ):
        partial_param_grids.append(
            {
                "scaler|ovs|clf__scaler__feature_range": [feature_range],
                "scaler|ovs|clf__ovs__k_neighbors": [k_neighbors],
                "scaler|ovs|clf__clf__pca__n_components": [n_components],
                "scaler|ovs|clf__clf__dtc__max_depth": [max_depth],
            }
        )
    exp_param_grids = []
    for rnd_seed, partial_param_grid in product(exp_random_states, partial_param_grids):
        partial_param_grid = partial_param_grid.copy()
        partial_param_grid.update(
            {
                "est_name": ["scaler|ovs|clf"],
                "scaler|ovs|clf__ovs__random_state": [rnd_seed],
                "scaler|ovs|clf__clf__dtc__random_state": [rnd_seed],
                "scaler|ovs|clf__clf__pca__random_state": [rnd_seed],
            }
        )
        exp_param_grids.append(partial_param_grid)

    # Assertions
    assert names[0] == exp_name
    assert steps[0] == exp_steps
    assert len(param_grids) == len(exp_param_grids)
    assert all([param_grid in exp_param_grids for param_grid in param_grids])


def test_check_pipelines_wrapper():
    """Based on the parameter keys error found in the experiment of a working paper."""

    # Initialization
    X, y = load_iris(return_X_y=True)
    n_runs = 1
    rnd_seed = 0
    oversamplers = [
        (
            "ovs",
            OverSamplingAugmentation(BorderlineSMOTE()),
            [
                {"oversampler__k_neighbors": [2, 4]},
                {"oversampler__m_neighbors": [6, 8]},
            ],
        )
    ]
    classifiers = [("clf", DecisionTreeClassifier(), {"max_depth": [3, 5]})]
    al_model = (
        "AL-TEST",
        StandardAL(max_iter=2),
        {"acquisition_func": ["random", "entropy", "breaking_ties"]},
    )

    we_wpg = check_pipelines_wrapper(
        classifiers,
        wrapper=al_model,
        random_state=rnd_seed,
        n_runs=n_runs,
        wrapped_only=True,
    )

    we_wpg2 = check_pipelines_wrapper(
        oversamplers,
        classifiers,
        wrapper=al_model,
        random_state=rnd_seed,
        n_runs=n_runs,
        wrapped_only=True,
    )
    for we, wpg in [we_wpg, we_wpg2]:
        ModelSearchCV(estimators=we, cv=2, param_grids=wpg, n_jobs=-1).fit(X, y)



import pytest
import numpy as np
import pandas as pd
from itertools import product
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import OrdinalEncoder, OneHotEncoder
from sklearn.tree import DecisionTreeClassifier

from .._encoders import PipelineEncoder


sklearn_encoders = [None, OrdinalEncoder(), OneHotEncoder()]


def generate_datasets():
    """
    Create datasets with different number of categorical features stored as both numpy
    and pandas formats.
    """

    MIN_CLASS = 20
    MAJ_CLASS = 100
    N_FEATURES = 4

    X = pd.DataFrame(
        np.random.random((30 + MIN_CLASS + MAJ_CLASS, N_FEATURES)),
        columns=[str(i) for i in range(N_FEATURES)],
    )
    y = pd.Series(
        [
            *[0 for i in range(MIN_CLASS)],
            *[1 for i in range(30)],
            *[2 for i in range(MAJ_CLASS)],
        ],
        name="target",
    )

    datasets = [(X, y, None), (X, y, [])]

    # Set up mixed type datasets with varying number of unique values
    for n_cat in range(1, N_FEATURES + 1):
        cat_columns = {
            col: f"cat_{i}"
            for i, col in enumerate(np.random.choice(X.columns, n_cat, replace=False))
        }
        X_ = X.rename(columns=cat_columns).copy()
        y_ = y.copy()

        for cat_col in cat_columns.values():
            X_.loc[:, cat_col] = (
                (X_.loc[:, cat_col] * np.random.randint(1, 10)).round().astype(int)
            )

        cat_cols_1 = X_.columns.str.startswith("cat_")
        cat_cols_2 = list(cat_cols_1)
        cat_cols_3 = list(np.where(cat_cols_1)[0])
        cat_cols_4 = X_.columns[cat_cols_1].tolist()

        # Pandas dataframes
        for cat_cols in [cat_cols_1, cat_cols_2, cat_cols_3, cat_cols_4]:
            datasets.append((X_.copy(), y_.copy(), cat_cols))

        # Numpy arrays
        for cat_cols in [cat_cols_1, cat_cols_2, cat_cols_3]:
            datasets.append((X_.copy().values, y_.copy().values, cat_cols))

        # Pandas + Numpy
        for cat_cols in [cat_cols_1, cat_cols_2, cat_cols_3, cat_cols_4]:
            datasets.append((X_.copy().values, y_.copy(), cat_cols))

        # Numpy + Pandas
        for cat_cols in [cat_cols_1, cat_cols_2, cat_cols_3]:
            datasets.append((X_.copy(), y_.copy().values, cat_cols))

    return datasets


@pytest.mark.parametrize("X, y, cat_features", generate_datasets())
def test_default_encoder(X, y, cat_features):

    encoder = PipelineEncoder(
        cat_features,
        sparse_output=False,
        dtype=np.float64,
    )

    if (
        cat_features is not None
        and len(cat_features) > 0
        and type(X) == np.ndarray
        and type(cat_features[0]) == str
    ):
        with pytest.raises(TypeError):
            encoder.fit_transform(X, y)
    else:
        X_ = encoder.fit_transform(X, y)
        if type(X) == pd.DataFrame:
            X = X.values

        if cat_features is None or len(cat_features) == 0:
            assert not encoder.features_.any()
            assert (X == X_).all()
        elif type(cat_features[0]) == np.bool_:
            assert encoder.features_.sum() == sum(cat_features)
        else:
            assert encoder.features_.sum() == len(cat_features)

        n_feats = sum([len(np.unique(feat)) for feat in X[:, encoder.features_].T])
        assert X_.shape == (X.shape[0], n_feats + sum(~encoder.features_))


@pytest.mark.parametrize(
    "sklearn_encoder, X, y, categorical_features",
    [
        (sklearn_encoder, *dataset)
        for sklearn_encoder, dataset in product(sklearn_encoders, generate_datasets())
    ],
)
def test_encoder(sklearn_encoder, X, y, categorical_features):

    # Check ordinal and One-Hot encoder
    encoder = PipelineEncoder(features=categorical_features, encoder=sklearn_encoder)

    if (
        categorical_features is not None
        and len(categorical_features) > 0
        and type(X) == np.ndarray
        and type(categorical_features[0]) == str
    ):
        with pytest.raises(TypeError):
            encoder.fit_transform(X, y)
    else:
        X_ = encoder.fit_transform(X, y)
        if type(X) == pd.DataFrame:
            X = X.values

        if sklearn_encoder.__class__.__name__ == "OrdinalEncoder":
            assert X.shape == X_.shape
        else:
            n_feats = sum([len(np.unique(feat)) for feat in X[:, encoder.features_].T])
            assert X_.shape == (
                X.shape[0],
                n_feats + sum(~encoder.features_),
            )


@pytest.mark.parametrize(
    "sklearn_encoder, X, y, categorical_features",
    [
        (sklearn_encoder, *dataset)
        for sklearn_encoder, dataset in product(sklearn_encoders, generate_datasets())
    ],
)
def test_pipeline_encoder(sklearn_encoder, X, y, categorical_features):

    # Check in Pipeline
    pipeline = make_pipeline(
        PipelineEncoder(features=categorical_features, encoder=sklearn_encoder),
        DecisionTreeClassifier(),
    )
    if (
        categorical_features is not None
        and len(categorical_features) > 0
        and type(X) == np.ndarray
        and type(categorical_features[0]) == str
    ):
        with pytest.raises(TypeError):
            pipeline.fit_transform(X, y)
    else:
        pipeline.fit(X, y)
        pipeline.predict(X)


def test_pipeline_encoder_errors():

    X, y, _ = generate_datasets()[3]

    err_cat_features = [[0, True, "cat_2"], [{}, {}, {}]]
    for cat_features in err_cat_features:
        encoder = PipelineEncoder(features=cat_features)
        with pytest.raises(TypeError):
            encoder.fit_transform(X, y)

    key_err_features = [["test"], ["test1", "test2"], "test"]
    for cat_features in key_err_features:
        encoder = PipelineEncoder(features=cat_features)
        with pytest.raises(KeyError):
            encoder.fit_transform(X, y)



"""
Download, transform and simulate various datasets.
"""

# Author: Joao Fonseca <jpfonseca@novaims.unl.pt>
# License: MIT

import io
import requests
import numpy as np
from scipy.io import loadmat

from .base import Datasets, FETCH_URLS
from ..utils import image_to_dataframe


class RemoteSensingDatasets(Datasets):
    """Class to download, transform and save remote sensing datasets."""

    def __init__(
        self,
        names: str = "all",
        data_home: str = None,
        download_if_missing: bool = True,
    ):
        """
        Parameters
        ----------
        names : str or list, default="all"
            List of dataset names to be downloaded. If ``all``, downloads all datasets.

        data_home : str, default=None
            The path to the data directory. If `None`, the default path
            is `~/ml_research_data`.

        download_if_missing : bool, default=True
            If True, downloads the dataset from the internet and puts it in
            ``data_home``. If the dataset is already downloaded, it is not downloaded
            again.

        Attributes
        ----------
        data_home_ : str
            Path were the data was stored.

        content_ : list
            List of tuples composed of (Dataset name, Dataframe).
        """
        self.names = names
        self.data_home = data_home
        self.download_if_missing = download_if_missing

    def download(self):
        """Download the datasets and append undersampled versions of them."""
        super(RemoteSensingDatasets, self).download(keep_index=True)
        content_ = []
        for (name, data) in self.content_:
            data = data.set_index(["h", "w"])
            content_.append((name, data))
        self.content_ = content_

        return self

    def _load_gic_dataset(self, dataset_name):
        for url in FETCH_URLS[dataset_name]:
            r = requests.get(url, stream=True)
            content = loadmat(io.BytesIO(r.content))
            arr = np.array(list(content.values())[-1])
            arr = np.expand_dims(arr, -1) if arr.ndim == 2 else arr
            if arr.shape[-1] != 1:
                indices = np.moveaxis(np.indices(arr.shape[:-1]), 0, -1)
                arr = np.insert(arr, [0, 0], indices, -1)
            yield arr

    def fetch_indian_pines(self):
        """Download and transform the Indian Pines Data Set. Label "0" means
        the pixel is not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Indian_Pines
        """
        df = image_to_dataframe(*self._load_gic_dataset("indian_pines"))
        return df[df.target != 0]

    def fetch_salinas(self):
        """Download and transform the Salinas Data Set. Label "0" means the pixel is
        not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Salinas_scene
        """
        df = image_to_dataframe(*self._load_gic_dataset("salinas"))
        return df[df.target != 0]

    def fetch_salinas_a(self):
        """Download and transform the Salinas-A Data Set. Label "0" means the pixel is
        not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Salinas-A_scene
        """
        df = image_to_dataframe(*self._load_gic_dataset("salinas_a"))
        return df[df.target != 0]

    def fetch_pavia_centre(self):
        """Download and transform the Pavia Centre Data Set. Label "0" means the pixel
        is not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Pavia_Centre_scene
        """
        df = image_to_dataframe(*self._load_gic_dataset("pavia_centre"))
        return df[df.target != 0]

    def fetch_pavia_university(self):
        """Download and transform the Pavia University Data Set. Label "0"
        means the pixel is not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Pavia_University_scene
        """
        df = image_to_dataframe(*self._load_gic_dataset("pavia_university"))
        return df[df.target != 0]

    def fetch_kennedy_space_center(self):
        """Download and transform the Kennedy Space Center Data Set. Label "0"
        means the pixel is not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Kennedy_Space_Center_.28KSC.29
        """
        df = image_to_dataframe(*self._load_gic_dataset("kennedy_space_center"))
        return df[df.target != 0]

    def fetch_botswana(self):
        """Download and transform the Botswana Data Set. Label "0" means the pixel is
        not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Botswana
        """
        df = image_to_dataframe(*self._load_gic_dataset("botswana"))
        return df[df.target != 0]



# Adapted from:
# https://github.com/scikit-learn-contrib/imbalanced-learn/blob/master/imblearn/tests/test_docstring_parameters.py

import inspect
import warnings
import importlib
from pkgutil import walk_packages
from inspect import signature

import pytest

from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.utils import IS_PYPY
from sklearn.utils._testing import check_docstring_parameters
from sklearn.utils._testing import _get_func_name
from sklearn.utils._testing import ignore_warnings
from sklearn.utils.estimator_checks import _enforce_estimator_tags_y
from sklearn.utils.estimator_checks import _enforce_estimator_tags_X
from sklearn.utils.estimator_checks import _construct_instance
from sklearn.utils.deprecation import _is_deprecated

import mlresearch
from mlresearch.utils._testing import all_estimators


def is_sampler(estimator):
    """Return True if the given estimator is a sampler, False otherwise.
    Parameters
    ----------
    estimator : object
        Estimator to test.
    Returns
    -------
    is_sampler : bool
        True if estimator is a sampler, otherwise False.
    """
    if estimator._estimator_type == "sampler":
        return True
    return False


# walk_packages() ignores DeprecationWarnings, now we need to ignore
# FutureWarnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore", FutureWarning)
    imblearn_path = mlresearch.__path__
    PUBLIC_MODULES = set(
        [
            pckg[1]
            for pckg in walk_packages(prefix="mlresearch.", path=imblearn_path)
            if not ("._" in pckg[1] or ".tests." in pckg[1])
        ]
    )

# functions to ignore args / docstring of
_DOCSTRING_IGNORES = []

# Methods where y param should be ignored if y=None by default
_METHODS_IGNORE_NONE_Y = [
    "fit",
    "score",
    "fit_predict",
    "fit_transform",
    "partial_fit",
    "predict",
]


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.skipif(IS_PYPY, reason="test segfaults on PyPy")
def test_docstring_parameters():
    # Test module docstring formatting

    # Skip test if numpydoc is not found
    pytest.importorskip(
        "numpydoc", reason="numpydoc is required to test the docstrings"
    )

    from numpydoc import docscrape

    incorrect = []
    for name in PUBLIC_MODULES:
        if name.endswith(".conftest"):
            # pytest tooling, not part of the scikit-learn API
            continue
        with warnings.catch_warnings(record=True):
            module = importlib.import_module(name)
        classes = inspect.getmembers(module, inspect.isclass)
        # Exclude non-scikit-learn classes
        classes = [cls for cls in classes if cls[1].__module__.startswith("mlresearch")]
        for cname, cls in classes:
            this_incorrect = []
            if cname in _DOCSTRING_IGNORES or cname.startswith("_"):
                continue
            if inspect.isabstract(cls):
                continue
            with warnings.catch_warnings(record=True) as w:
                cdoc = docscrape.ClassDoc(cls)
            if len(w):
                raise RuntimeError(
                    "Error for __init__ of %s in %s:\n%s" % (cls, name, w[0])
                )

            cls_init = getattr(cls, "__init__", None)

            if _is_deprecated(cls_init):
                continue
            elif cls_init is not None:
                this_incorrect += check_docstring_parameters(cls.__init__, cdoc)

            for method_name in cdoc.methods:
                method = getattr(cls, method_name)
                if _is_deprecated(method):
                    continue
                param_ignore = None
                # Now skip docstring test for y when y is None
                # by default for API reason
                if method_name in _METHODS_IGNORE_NONE_Y:
                    sig = signature(method)
                    if "y" in sig.parameters and sig.parameters["y"].default is None:
                        param_ignore = ["y"]  # ignore y for fit and score
                result = check_docstring_parameters(method, ignore=param_ignore)
                this_incorrect += result

            incorrect += this_incorrect

        functions = inspect.getmembers(module, inspect.isfunction)
        # Exclude imported functions
        functions = [fn for fn in functions if fn[1].__module__ == name]
        for fname, func in functions:
            # Don't test private methods / functions
            if fname.startswith("_"):
                continue
            if fname == "configuration" and name.endswith("setup"):
                continue
            name_ = _get_func_name(func)
            if not any(d in name_ for d in _DOCSTRING_IGNORES) and not _is_deprecated(
                func
            ):
                incorrect += check_docstring_parameters(func)

    msg = "\n".join(incorrect)
    if len(incorrect) > 0:
        raise AssertionError("Docstring Error:\n" + msg)


@ignore_warnings(category=FutureWarning)
def test_tabs():
    # Test that there are no tabs in our source files
    for importer, modname, ispkg in walk_packages(
        mlresearch.__path__, prefix="mlresearch."
    ):

        if IS_PYPY:
            continue

        # because we don't import
        mod = importlib.import_module(modname)

        try:
            source = inspect.getsource(mod)
        except IOError:  # user probably should have run "make clean"
            continue
        assert "\t" not in source, (
            '"%s" has tabs, please remove them ',
            "or add it to the ignore list" % modname,
        )


def _construct_compose_pipeline_instance(Estimator):
    # Minimal / degenerate instances: only useful to test the docstrings.
    if Estimator.__name__ == "Pipeline":
        return Estimator(steps=[("clf", LogisticRegression())])


@pytest.mark.parametrize("name, Estimator", all_estimators())
def test_fit_docstring_attributes(name, Estimator):
    pytest.importorskip("numpydoc")
    from numpydoc import docscrape

    if Estimator.__name__ in _DOCSTRING_IGNORES:
        return

    doc = docscrape.ClassDoc(Estimator)
    attributes = doc["Attributes"]

    if Estimator.__name__ == "Pipeline":
        est = _construct_compose_pipeline_instance(Estimator)
    else:
        est = _construct_instance(Estimator)

    X, y = make_classification(
        n_samples=20,
        n_features=3,
        n_redundant=0,
        n_classes=2,
        random_state=2,
    )

    y = _enforce_estimator_tags_y(est, y)
    X = _enforce_estimator_tags_X(est, X)

    if "oob_score" in est.get_params():
        est.set_params(oob_score=True)

    if is_sampler(est):
        est.fit_resample(X, y)
    else:
        est.fit(X, y)

    skipped_attributes = set([])

    for attr in attributes:
        if attr.name in skipped_attributes:
            continue
        desc = " ".join(attr.desc).lower()
        # As certain attributes are present "only" if a certain parameter is
        # provided, this checks if the word "only" is present in the attribute
        # description, and if not the attribute is required to be present.
        if "only " in desc:
            continue
        # ignore deprecation warnings
        with ignore_warnings(category=FutureWarning):
            assert hasattr(est, attr.name)

    fit_attr = _get_all_fitted_attributes(est)
    fit_attr_names = [attr.name for attr in attributes]
    undocumented_attrs = set(fit_attr).difference(fit_attr_names)
    undocumented_attrs = set(undocumented_attrs).difference(skipped_attributes)
    if undocumented_attrs:
        raise AssertionError(
            f"Undocumented attributes for {Estimator.__name__}: {undocumented_attrs}"
        )


def _get_all_fitted_attributes(estimator):
    "Get all the fitted attributes of an estimator including properties"
    # attributes
    fit_attr = list(estimator.__dict__.keys())

    # properties
    with warnings.catch_warnings():
        warnings.filterwarnings("error", category=FutureWarning)

        for name in dir(estimator.__class__):
            obj = getattr(estimator.__class__, name)
            if not isinstance(obj, property):
                continue

            # ignore properties that raises an AttributeError and deprecated
            # properties
            try:
                getattr(estimator, name)
            except (AttributeError, FutureWarning):
                continue
            fit_attr.append(name)

    return [k for k in fit_attr if k.endswith("_") and not k.startswith("_")]



"""
Data I/O utils. Later on I might add other data handling utilities.
"""
from os import listdir
from os.path import isdir, join
import pandas as pd
from sqlite3 import connect


def load_datasets(
    data_dir, prefix="", suffix="", target_exists=True, **read_csv_kwargs
):
    """
    Load all datasets in a directory from sqlite databases and/or csv files.

    Parameters
    ----------
    data_dir : str
        Data directory to be crawled.

    prefix : str, default=''
        Load dataset if the file starts with the specified prefix.

    suffix : str, default=''
        Load dataset if the file starts with the specified suffix.

    target_exists : bool, default=True
        Specify wether there is a target feature. If True, it is assumed to be in the
        last position of the dataset.

    Returns
    -------
    datasets : list
        A list with nested tuples with structure (dataset_name, (X, y)).
    """
    assert isdir(data_dir), "`data_dir` must be a directory."

    # Filter data by suffix
    dat_names = [
        dat
        for dat in listdir(data_dir)
        if (dat.startswith(prefix) and dat.endswith(suffix))
    ]

    # Read data
    datasets = []
    for dat_name in dat_names:
        data_path = join(data_dir, dat_name)

        # Handle csv data
        if dat_name.endswith(".csv"):
            ds = pd.read_csv(data_path, **read_csv_kwargs)
            name = dat_name.replace(".csv", "").replace("_", " ").upper()
            if target_exists:
                ds = (ds.iloc[:, :-1], ds.iloc[:, -1])
            datasets.append((name, ds))

        # Handle sqlite database
        elif dat_name.endswith(".db"):
            with connect(data_path) as connection:
                datasets_names = [
                    name[0]
                    for name in connection.execute(
                        "SELECT name FROM sqlite_master WHERE type='table';"
                    )
                ]
                for dataset_name in datasets_names:
                    ds = pd.read_sql(f'select * from "{dataset_name}"', connection)
                    if target_exists:
                        ds = (ds.iloc[:, :-1], ds.iloc[:, -1])
                    datasets.append((dataset_name.replace("_", " ").upper(), ds))
    return datasets



"""
Test the geometric_smote module.
"""

from collections import Counter

import pytest
import numpy as np
from numpy.linalg import norm
from sklearn.utils import check_random_state
from sklearn.utils._testing import assert_allclose, assert_array_equal
from sklearn.datasets import make_classification
from scipy import sparse

from .._gsmote import _make_geometric_sample, GeometricSMOTE, SELECTION_STRATEGY

RND_SEED = 0
RANDOM_STATE = check_random_state(RND_SEED)
CENTERS = [
    RANDOM_STATE.random_sample((2,)),
    2.6 * RANDOM_STATE.random_sample((4,)),
    3.2 * RANDOM_STATE.random_sample((10,)),
    -0.5 * RANDOM_STATE.random_sample((1,)),
]
SURFACE_POINTS = [
    RANDOM_STATE.random_sample((2,)),
    5.2 * RANDOM_STATE.random_sample((4,)),
    -3.5 * RANDOM_STATE.random_sample((10,)),
    -10.9 * RANDOM_STATE.random_sample((1,)),
]
TRUNCATION_FACTORS = [-1.0, -0.5, 0.0, 0.5, 1.0]
DEFORMATION_FACTORS = [0.0, 0.25, 0.5, 0.75, 1.0]


def data_heterogeneous_ordered():
    rng = np.random.RandomState(42)
    X = np.empty((30, 4), dtype=object)
    # create 2 random continuous feature
    X[:, :2] = rng.randn(30, 2)
    # create a categorical feature using some string
    X[:, 2] = rng.choice(["a", "b", "c"], size=30).astype(object)
    # create a categorical feature using some integer
    X[:, 3] = rng.randint(3, size=30)
    y = np.array([0] * 10 + [1] * 20)
    # return the categories
    return X, y, [2, 3]


def data_heterogeneous_unordered():
    rng = np.random.RandomState(42)
    X = np.empty((30, 4), dtype=object)
    # create 2 random continuous feature
    X[:, [1, 2]] = rng.randn(30, 2)
    # create a categorical feature using some string
    X[:, 0] = rng.choice(["a", "b", "c"], size=30).astype(object)
    # create a categorical feature using some integer
    X[:, 3] = rng.randint(3, size=30)
    y = np.array([0] * 10 + [1] * 20)
    # return the categories
    return X, y, [0, 3]


def data_heterogeneous_masked():
    rng = np.random.RandomState(42)
    X = np.empty((30, 4), dtype=object)
    # create 2 random continuous feature
    X[:, [1, 2]] = rng.randn(30, 2)
    # create a categorical feature using some string
    X[:, 0] = rng.choice(["a", "b", "c"], size=30).astype(object)
    # create a categorical feature using some integer
    X[:, 3] = rng.randint(3, size=30)
    y = np.array([0] * 10 + [1] * 20)
    # return the categories
    return X, y, [True, False, True]


def data_heterogeneous_unordered_multiclass():
    rng = np.random.RandomState(42)
    X = np.empty((50, 4), dtype=object)
    # create 2 random continuous feature
    X[:, [1, 2]] = rng.randn(50, 2)
    # create a categorical feature using some string
    X[:, 0] = rng.choice(["a", "b", "c"], size=50).astype(object)
    # create a categorical feature using some integer
    X[:, 3] = rng.randint(3, size=50)
    y = np.array([0] * 10 + [1] * 15 + [2] * 25)
    # return the categories
    return X, y, [0, 3]


def data_sparse(format):
    rng = np.random.RandomState(42)
    X = np.empty((30, 4), dtype=np.float64)
    # create 2 random continuous feature
    X[:, [1, 2]] = rng.randn(30, 2)
    # create a categorical feature using some string
    X[:, 0] = rng.randint(3, size=30)
    # create a categorical feature using some integer
    X[:, 3] = rng.randint(3, size=30)
    y = np.array([0] * 10 + [1] * 20)
    X = sparse.csr_matrix(X) if format == "csr" else sparse.csc_matrix(X)
    return X, y, [0, 3]


@pytest.mark.parametrize(
    "center,surface_point",
    [
        (CENTERS[0], SURFACE_POINTS[0]),
        (CENTERS[1], SURFACE_POINTS[1]),
        (CENTERS[2], SURFACE_POINTS[2]),
        (CENTERS[3], SURFACE_POINTS[3]),
    ],
)
def test_make_geometric_sample_hypersphere(center, surface_point):
    """Test the generation of points inside a hypersphere."""
    point = _make_geometric_sample(center, surface_point, 0.0, 0.0, RANDOM_STATE)
    rel_point = point - center
    rel_surface_point = surface_point - center
    np.testing.assert_array_less(0.0, norm(rel_surface_point) - norm(rel_point))


@pytest.mark.parametrize(
    "surface_point,deformation_factor",
    [
        (np.array([1.0, 0.0]), 0.0),
        (2.6 * np.array([0.0, 1.0]), 0.25),
        (3.2 * np.array([0.0, 1.0, 0.0, 0.0]), 0.50),
        (0.5 * np.array([0.0, 0.0, 1.0]), 0.75),
        (6.7 * np.array([0.0, 0.0, 1.0, 0.0, 0.0]), 1.0),
    ],
)
def test_make_geometric_sample_half_hypersphere(surface_point, deformation_factor):
    """Test the generation of points inside a hypersphere."""
    center = np.zeros(surface_point.shape)
    point = _make_geometric_sample(
        center, surface_point, 1.0, deformation_factor, RANDOM_STATE
    )
    np.testing.assert_array_less(0.0, norm(surface_point) - norm(point))
    np.testing.assert_array_less(0.0, np.dot(point, surface_point))


@pytest.mark.parametrize(
    "center,surface_point,truncation_factor",
    [
        (center, surface_point, truncation_factor)
        for center, surface_point in zip(CENTERS, SURFACE_POINTS)
        for truncation_factor in TRUNCATION_FACTORS
    ],
)
def test_make_geometric_sample_line_segment(center, surface_point, truncation_factor):
    """Test the generation of points on a line segment."""
    point = _make_geometric_sample(
        center, surface_point, truncation_factor, 1.0, RANDOM_STATE
    )
    rel_point = point - center
    rel_surface_point = surface_point - center
    dot_product = np.dot(rel_point, rel_surface_point)
    norms_product = norm(rel_point) * norm(rel_surface_point)
    np.testing.assert_array_less(0.0, norm(rel_surface_point) - norm(rel_point))
    dot_product = (
        np.abs(dot_product) if truncation_factor == 0.0 else (-1) * dot_product
    )
    np.testing.assert_allclose(np.abs(dot_product) / norms_product, 1.0)


def test_gsmote_default_init():
    """Test the intialization with default parameters."""
    gsmote = GeometricSMOTE()
    assert gsmote.sampling_strategy == "auto"
    assert gsmote.random_state is None
    assert gsmote.truncation_factor == 1.0
    assert gsmote.deformation_factor == 0.0
    assert gsmote.selection_strategy == "combined"
    assert gsmote.k_neighbors == 5
    assert gsmote.categorical_features is None
    assert gsmote.n_jobs == 1


def test_gsmote_fit():
    """Test fit method."""
    n_samples, weights = 200, [0.6, 0.4]
    X, y = make_classification(
        random_state=RND_SEED, n_samples=n_samples, weights=weights
    )
    gsmote = GeometricSMOTE(random_state=RANDOM_STATE).fit(X, y)
    assert gsmote.sampling_strategy_ == {1: 40}


def test_gsmote_invalid_selection_strategy():
    """Test invalid selection strategy."""
    n_samples, weights = 200, [0.6, 0.4]
    X, y = make_classification(
        random_state=RND_SEED, n_samples=n_samples, weights=weights
    )
    gsmote = GeometricSMOTE(random_state=RANDOM_STATE, selection_strategy="Minority")
    with pytest.raises(ValueError):
        gsmote.fit_resample(X, y)


@pytest.mark.parametrize("selection_strategy", ["combined", "minority", "majority"])
def test_gsmote_nn(selection_strategy):
    """Test nearest neighbors object."""
    n_samples, weights = 200, [0.6, 0.4]
    X, y = make_classification(
        random_state=RND_SEED, n_samples=n_samples, weights=weights
    )
    gsmote = GeometricSMOTE(
        random_state=RANDOM_STATE, selection_strategy=selection_strategy
    )
    _ = gsmote.fit_resample(X, y)
    if selection_strategy in ("minority", "combined"):
        assert gsmote.nns_pos_.n_neighbors == gsmote.k_neighbors + 1
    if selection_strategy in ("majority", "combined"):
        assert gsmote.nn_neg_.n_neighbors == 1


@pytest.mark.parametrize(
    "selection_strategy, truncation_factor, deformation_factor",
    [
        (selection_strategy, truncation_factor, deformation_factor)
        for selection_strategy in SELECTION_STRATEGY
        for truncation_factor in TRUNCATION_FACTORS
        for deformation_factor in DEFORMATION_FACTORS
    ],
)
def test_gsmote_fit_resample_binary(
    selection_strategy, truncation_factor, deformation_factor
):
    """Test fit and sample for binary class case."""
    n_maj, n_min, step, min_coor, max_coor = 12, 5, 0.5, 0.0, 8.5
    X = np.repeat(np.arange(min_coor, max_coor, step), 2).reshape(-1, 2)
    y = np.concatenate([np.repeat(0, n_maj), np.repeat(1, n_min)])
    radius = np.sqrt(0.5) * step
    k_neighbors = 1
    gsmote = GeometricSMOTE(
        "auto",
        RANDOM_STATE,
        truncation_factor,
        deformation_factor,
        selection_strategy,
        k_neighbors,
    )
    X_resampled, y_resampled = gsmote.fit_resample(X, y)
    assert gsmote.sampling_strategy_ == {1: (n_maj - n_min)}
    assert y_resampled.sum() == n_maj
    np.testing.assert_array_less(X[n_maj - 1] - radius, X_resampled[n_maj + n_min])


@pytest.mark.parametrize(
    "selection_strategy, truncation_factor, deformation_factor",
    [
        (selection_strategy, truncation_factor, deformation_factor)
        for selection_strategy in SELECTION_STRATEGY
        for truncation_factor in TRUNCATION_FACTORS
        for deformation_factor in DEFORMATION_FACTORS
    ],
)
def test_gsmote_fit_resample_multiclass(
    selection_strategy, truncation_factor, deformation_factor
):
    """Test fit and sample for multiclass case."""
    n_samples, weights = 100, [0.75, 0.15, 0.10]
    X, y = make_classification(
        random_state=RND_SEED,
        n_samples=n_samples,
        weights=weights,
        n_classes=3,
        n_informative=5,
    )
    k_neighbors, majority_label = 1, 0
    gsmote = GeometricSMOTE(
        "auto",
        RANDOM_STATE,
        truncation_factor,
        deformation_factor,
        selection_strategy,
        k_neighbors,
    )
    _, y_resampled = gsmote.fit_resample(X, y)
    assert majority_label not in gsmote.sampling_strategy_.keys()
    np.testing.assert_array_equal(np.unique(y), np.unique(y_resampled))
    assert len(set(Counter(y_resampled).values())) == 1


def test_categorical_error():
    X, y, _ = data_heterogeneous_unordered()
    categorical_features = [0, 10]
    gsmote = GeometricSMOTE(random_state=0, categorical_features=categorical_features)
    with pytest.raises(ValueError, match="indices are out of range"):
        gsmote.fit_resample(X, y)


@pytest.mark.parametrize(
    "data",
    [
        data_heterogeneous_ordered(),
        data_heterogeneous_unordered(),
        data_heterogeneous_masked(),
        data_sparse("csr"),
        data_sparse("csc"),
    ],
)
def test_gsmotenc(data):
    X, y, categorical_features = data
    gsmote = GeometricSMOTE(random_state=0, categorical_features=categorical_features)
    X_resampled, y_resampled = gsmote.fit_resample(X, y)

    assert X_resampled.dtype == X.dtype

    categorical_features = np.array(categorical_features)
    if categorical_features.dtype == bool:
        categorical_features = np.flatnonzero(categorical_features)
    for cat_idx in categorical_features:
        if sparse.issparse(X):
            assert set(X[:, cat_idx].data) == set(X_resampled[:, cat_idx].data)
            assert X[:, cat_idx].dtype == X_resampled[:, cat_idx].dtype
        else:
            assert set(X[:, cat_idx]) == set(X_resampled[:, cat_idx])
            assert X[:, cat_idx].dtype == X_resampled[:, cat_idx].dtype


# part of the common test which apply to GSMOTE-NC even if it is not default
# constructible
def test_smotenc_check_target_type():
    X, _, categorical_features = data_heterogeneous_unordered()
    y = np.linspace(0, 1, 30)
    gsmote = GeometricSMOTE(categorical_features=categorical_features, random_state=0)
    with pytest.raises(ValueError, match="Unknown label type: 'continuous'"):
        gsmote.fit_resample(X, y)
    rng = np.random.RandomState(42)
    y = rng.randint(2, size=(20, 3))
    msg = "Multilabel and multioutput targets are not supported."
    with pytest.raises(ValueError, match=msg):
        gsmote.fit_resample(X, y)


def test_smotenc_samplers_one_label():
    X, _, categorical_features = data_heterogeneous_unordered()
    y = np.zeros(30)
    gsmote = GeometricSMOTE(categorical_features=categorical_features, random_state=0)
    with pytest.raises(ValueError, match="needs to have more than 1 class"):
        gsmote.fit(X, y)


def test_smotenc_fit():
    X, y, categorical_features = data_heterogeneous_unordered()
    gsmote = GeometricSMOTE(categorical_features=categorical_features, random_state=0)
    gsmote.fit_resample(X, y)
    assert hasattr(
        gsmote, "sampling_strategy_"
    ), "No fitted attribute sampling_strategy_"


def test_smotenc_fit_resample():
    X, y, categorical_features = data_heterogeneous_unordered()
    target_stats = Counter(y)
    gsmote = GeometricSMOTE(categorical_features=categorical_features, random_state=0)
    _, y_res = gsmote.fit_resample(X, y)
    _ = Counter(y_res)
    n_samples = max(target_stats.values())
    assert all(value >= n_samples for value in Counter(y_res).values())


def test_smotenc_fit_resample_sampling_strategy():
    X, y, categorical_features = data_heterogeneous_unordered_multiclass()
    expected_stat = Counter(y)[1]
    gsmote = GeometricSMOTE(categorical_features=categorical_features, random_state=0)
    sampling_strategy = {2: 25, 0: 25}
    gsmote.set_params(sampling_strategy=sampling_strategy)
    X_res, y_res = gsmote.fit_resample(X, y)
    assert Counter(y_res)[1] == expected_stat


def test_smotenc_pandas():
    pd = pytest.importorskip("pandas")
    # Check that the samplers handle pandas dataframe and pandas series
    X, y, categorical_features = data_heterogeneous_unordered_multiclass()
    X_pd = pd.DataFrame(X)
    gsmote = GeometricSMOTE(categorical_features=categorical_features, random_state=0)
    X_res_pd, y_res_pd = gsmote.fit_resample(X_pd, y)
    X_res, y_res = gsmote.fit_resample(X, y)
    assert_array_equal(X_res_pd.to_numpy(), X_res)
    assert_allclose(y_res_pd, y_res)


def test_smotenc_preserve_dtype():
    X, y = make_classification(
        n_samples=50,
        n_classes=3,
        n_informative=4,
        weights=[0.2, 0.3, 0.5],
        random_state=0,
    )
    # Cast X and y to not default dtype
    X = X.astype(np.float32)
    y = y.astype(np.int32)
    gsmote = GeometricSMOTE(categorical_features=[1], random_state=0)
    X_res, y_res = gsmote.fit_resample(X, y)
    assert X.dtype == X_res.dtype, "X dtype is not preserved"
    assert y.dtype == y_res.dtype, "y dtype is not preserved"


@pytest.mark.parametrize("categorical_features", [[True, True, True], [0, 1, 2]])
def test_smotenc_raising_error_all_categorical(categorical_features):
    X, y = make_classification(
        n_features=3,
        n_informative=1,
        n_redundant=1,
        n_repeated=0,
        n_clusters_per_class=1,
    )
    gsmote = GeometricSMOTE(categorical_features=categorical_features)
    err_msg = "GeometricSMOTE is not designed to work only with categorical features"
    with pytest.raises(ValueError, match=err_msg):
        gsmote.fit_resample(X, y)


def test_smote_nc_with_null_median_std():
    # Non-regression test for #662
    # https://github.com/scikit-learn-contrib/imbalanced-learn/issues/662
    data = np.array(
        [
            [1, 2, 1, "A"],
            [2, 1, 2, "A"],
            [1, 2, 3, "B"],
            [1, 2, 4, "C"],
            [1, 2, 5, "C"],
        ],
        dtype="object",
    )
    labels = np.array(
        ["class_1", "class_1", "class_1", "class_2", "class_2"], dtype=object
    )
    gsmote = GeometricSMOTE(
        categorical_features=[3],
        k_neighbors=1,
        selection_strategy="minority",
        random_state=0,
    )
    X_res, y_res = gsmote.fit_resample(data, labels)
    # check that the categorical feature is not random but correspond to the
    # categories seen in the minority class samples
    assert X_res[-1, -1] == "C"



import pytest
import re
import numpy as np
from collections import Counter
from sklearn.neighbors import NearestNeighbors
from sklearn.utils._testing import ignore_warnings
from imblearn.over_sampling import (
    RandomOverSampler,
    SMOTE,
    BorderlineSMOTE,
    SVMSMOTE,
)

from .._oversampling_augmentation import (
    _modify_nn,
    _clone_modify,
    OverSamplingAugmentation,
)

RANDOM_STATE = 42
OVERSAMPLERS = [
    RandomOverSampler(random_state=RANDOM_STATE),
    SMOTE(random_state=RANDOM_STATE),
    BorderlineSMOTE(random_state=RANDOM_STATE),
    SVMSMOTE(random_state=RANDOM_STATE),
]

X = np.array([(0.0, 0.0), (1.0, 1.0), (2.0, 2.0), (3.0, 3.0), (4.0, 4.0)])
y = np.array([0, 0, 1, 1, 1])


def test_modify_nn_object():
    """Test the modification of nn object."""
    assert _modify_nn(NearestNeighbors(n_neighbors=5), 3).n_neighbors == 2
    assert _modify_nn(NearestNeighbors(n_neighbors=3), 3).n_neighbors == 2
    assert _modify_nn(NearestNeighbors(n_neighbors=2), 5).n_neighbors == 2


def test_modify_nn_int():
    """Test the modification of integer nn."""
    assert _modify_nn(5, 3) == 2
    assert _modify_nn(3, 3) == 2
    assert _modify_nn(2, 5) == 2


def test_clone_modify_ros():
    """Test the cloning and modification of random oversampler."""
    cloned_oversampler = _clone_modify(OVERSAMPLERS[0], None)
    assert isinstance(cloned_oversampler, RandomOverSampler)


@pytest.mark.parametrize(
    "oversampler",
    [ovs for ovs in OVERSAMPLERS if not isinstance(ovs, RandomOverSampler)],
)
def test_clone_modify_single_min_sample(oversampler):
    """Test the cloning and modification for one minority class sample."""
    y = np.array([0, 0, 0, 0, 1, 2, 2, 2])
    cloned_oversampler = _clone_modify(oversampler, y)
    assert isinstance(cloned_oversampler, RandomOverSampler)


@pytest.mark.parametrize(
    "oversampler",
    [ovs for ovs in OVERSAMPLERS if not isinstance(ovs, RandomOverSampler)],
)
def test_clone_modify_neighbors(oversampler):
    """Test the cloning and modification of neighbors based oversamplers."""
    y = np.array([0, 0, 0, 0, 2, 2, 2])
    n_minority_samples = Counter(y).most_common()[-1][1]
    cloned_oversampler = _clone_modify(oversampler, y)
    assert isinstance(cloned_oversampler, oversampler.__class__)
    if hasattr(cloned_oversampler, "k_neighbors"):
        assert cloned_oversampler.k_neighbors == n_minority_samples - 1
    if hasattr(cloned_oversampler, "m_neighbors"):
        assert (cloned_oversampler.m_neighbors == y.size - 1) or (
            cloned_oversampler.m_neighbors == "deprecated"
        )
    if hasattr(cloned_oversampler, "n_neighbors"):
        assert (cloned_oversampler.n_neighbors == n_minority_samples - 1) or (
            cloned_oversampler.n_neighbors == "deprecated"
        )


@pytest.mark.parametrize(
    "generator",
    [
        OverSamplingAugmentation(
            oversampler=SMOTE(k_neighbors=5),
            random_state=RANDOM_STATE,
        ),
        OverSamplingAugmentation(
            oversampler=SMOTE(k_neighbors=5, random_state=RANDOM_STATE),
            random_state=RANDOM_STATE,
            augmentation_strategy="constant",
            value=10,
        ),
        OverSamplingAugmentation(
            oversampler=SMOTE(k_neighbors=5, random_state=RANDOM_STATE),
            random_state=RANDOM_STATE,
            augmentation_strategy="proportional",
            value=10,
        ),
        OverSamplingAugmentation(
            oversampler=SMOTE(k_neighbors=5, random_state=RANDOM_STATE),
            random_state=RANDOM_STATE,
            augmentation_strategy=2,
        ),
        OverSamplingAugmentation(
            oversampler=SMOTE(k_neighbors=5, random_state=RANDOM_STATE),
            random_state=RANDOM_STATE,
            augmentation_strategy={0: 6, 1: 10},
        ),
    ],
)
@ignore_warnings
def test_fit_resample(generator):
    """Test the fit_resample method for various
    cases and data generator."""
    n_exp_obs = {
        "oversampling": {0: 3, 1: 3},
        "constant": {0: 10, 1: 10},
        "proportional": {0: 4, 1: 6},
        "2": {0: 4, 1: 6},
        "{0: 6, 1: 10}": {0: 6, 1: 10},
    }

    X_res, y_res = generator.fit_resample(X, y)
    y_count = dict(Counter(y_res))
    assert y_count == n_exp_obs[str(generator.augmentation_strategy)]
    assert generator._fit_resample(X, y) is None  # _fit_resample shouldn't do anything


@pytest.mark.parametrize(
    "generator, err_msg",
    [
        (
            OverSamplingAugmentation(
                oversampler=SMOTE(k_neighbors=5),
                augmentation_strategy="proportional",
                value=2,
            ),
            (
                "The new size of the augmented dataset must be larger than the original "
                "dataset. Originally, there are 5 samples and 2 samples are asked."
            ),
        ),
        (
            OverSamplingAugmentation(
                oversampler=SMOTE(k_neighbors=5),
                augmentation_strategy="option",
            ),
            (
                "When 'augmentation_strategy' is neither an int or float, it needs to be"
                " one of ['oversampling', 'constant', 'proportional']. Got 'option' "
                "instead."
            ),
        ),
        (
            OverSamplingAugmentation(
                oversampler=SMOTE(k_neighbors=5),
                augmentation_strategy="constant",
                value="10",
            ),
            (
                "When 'augmentation_strategy' is 'constant' or 'proportional', 'value' "
                "needs to be an int or float. Got 10 instead."
            ),
        ),
        (
            OverSamplingAugmentation(
                oversampler=SMOTE(k_neighbors=5),
                augmentation_strategy="proportional",
                value="10",
            ),
            (
                "When 'augmentation_strategy' is 'constant' or 'proportional', 'value' "
                "needs to be an int or float. Got 10 instead."
            ),
        ),
    ],
)
def test_errors(generator, err_msg):
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        generator.fit_resample(X, y)



"""
Test conversions between image and dataframe.
"""
from itertools import product
import pytest
import numpy as np

from .._image import image_to_dataframe, dataframe_to_image

RND_SEED = 42


@pytest.mark.parametrize("number_of_bands", list(range(1, 20, 3)))
def test_image_to_dataframe(number_of_bands):
    """Test the conversion of image arrays to regular dataframes."""
    shape = (150, 200, number_of_bands)
    band_names = [f"a{i}" for i in range(number_of_bands)]

    X = np.random.default_rng(RND_SEED).random(shape)
    target = (
        np.random.default_rng(RND_SEED)
        .integers(0, 2, shape[0] * shape[1])
        .reshape(shape[0], shape[1])
    )

    for bands, y in product([band_names, None], [target, None]):
        df = image_to_dataframe(X=X, y=y, bands=bands)

        if y is None:
            assert df.shape == (shape[0] * shape[1], shape[2])
            for h, w in product([0, shape[0] - 1], [0, shape[1] - 1]):
                assert (df.loc[h, w] == X[h, w]).all()
        else:
            assert df.shape == (shape[0] * shape[1], shape[2] + 1)
            assert df.columns[-1] == "target"
            assert df["target"].dtype == int
            for h, w in product([0, shape[0] - 1], [0, shape[1] - 1]):
                assert (df.loc[h, w] == np.append(X[h, w], y[h, w])).all()

        if bands is None:
            assert df.columns[:number_of_bands].tolist() == list(range(number_of_bands))
        else:
            assert df.columns[:number_of_bands].tolist() == band_names


@pytest.mark.parametrize("number_of_bands", list(range(1, 20, 3)))
def test_dataframe_to_image(number_of_bands):

    shape = (150, 200, number_of_bands)
    band_names = [f"a{i}" for i in range(number_of_bands)]

    X = np.random.default_rng(RND_SEED).random(shape)
    target = (
        np.random.default_rng(RND_SEED)
        .integers(0, 2, shape[0] * shape[1])
        .reshape(shape[0], shape[1])
    )

    for bands, y in product([band_names, None], [target, None]):
        df = image_to_dataframe(X=X, y=y, bands=bands)
        target_ = "target" if y is not None else None
        X_, y_, bands_ = dataframe_to_image(df=df, target_feature=target_)

        if y is None:
            assert y_ is None
        else:
            assert type(y_) == np.ndarray
            assert (y_ == y).all()

        assert (bands_ == df.columns.drop(target_, errors="ignore")).all()
        assert type(X_) == np.ndarray
        assert (X_ == X).all()

        # Without "h", "w" features at all
        with pytest.raises(IndexError):
            dataframe_to_image(df=df.reset_index(drop=True), target_feature=target_)

        # With "h", "w" in columns
        with pytest.raises(IndexError):
            dataframe_to_image(df=df.reset_index(drop=True), target_feature=target_)

    # Test passing a subset of bands
    if len(band_names) > 2:
        df = image_to_dataframe(X=X, y=target, bands=band_names)
        bands_filter = np.random.choice(band_names, 2, replace=False)

        X_, y_, bands_ = dataframe_to_image(
            df=df, bands=bands_filter, target_feature="target"
        )

        assert (bands_ == bands_filter).all()
        assert type(y_) == np.ndarray
        assert type(X_) == np.ndarray
        assert (y_ == target).all()

        band_indices = [
            np.where(df.columns.drop("target") == b)[0][0] for b in bands_filter
        ]
        assert (X_ == X[:, :, band_indices]).all()



"""
This module contains various performance metrics/scorers that are not
included in scikit-learn's scorers' dictionary. Additionally, an expanded
dictionary of scorers (as compared with scikit-learn's) is also provided.
"""
from ._metrics import (
    geometric_mean_score_macro,
    ALScorer,
    area_under_learning_curve,
    data_utilization_rate,
)

# Update the list of scorers
from sklearn.metrics import get_scorer, get_scorer_names

__all__ = [
    "get_scorer",
    "get_scorer_names",
    "geometric_mean_score_macro",
    "area_under_learning_curve",
    "data_utilization_rate",
    "ALScorer",
]



"""
This module contains a variety of general utility functions and tools used to format and
prepare tables to incorporate into LaTeX code.
"""
from ._image import image_to_dataframe, dataframe_to_image
from ._data import load_datasets
from ._check_pipelines import check_pipelines, check_pipelines_wrapper
from ._visualization import set_matplotlib_style, feature_to_color
from ._utils import generate_paths

__all__ = [
    "image_to_dataframe",
    "dataframe_to_image",
    "load_datasets",
    "check_pipelines",
    "check_pipelines_wrapper",
    "set_matplotlib_style",
    "feature_to_color",
    "generate_paths",
]



"""
Module which contains Active Learning implementations.
"""
from ._active_learning import StandardAL, AugmentationAL
from ._acquisition_functions import ACQUISITION_FUNCTIONS

__all__ = ["StandardAL", "AugmentationAL", "ACQUISITION_FUNCTIONS"]



"""
A wrapper to allow the use of oversampling algorithms for data augmentation
in Active Learning experiments with multiple datasets.
"""
import warnings
from collections import Counter, OrderedDict
import numpy as np
from sklearn.base import clone
from sklearn.neighbors import NearestNeighbors
from imblearn.over_sampling.base import BaseOverSampler
from imblearn.over_sampling import RandomOverSampler
from ._gsmote import GeometricSMOTE

AUGMENTATION_STRATEGIES = ["oversampling", "constant", "proportional"]


def _modify_nn(n_neighbors, n_samples):
    """Modify nearest neighbors object or integer."""
    if isinstance(n_neighbors, NearestNeighbors):
        n_neighbors = (
            clone(n_neighbors).set_params(n_neighbors=n_samples - 1)
            if n_neighbors.n_neighbors >= n_samples
            else clone(n_neighbors)
        )
    elif isinstance(n_neighbors, int) and n_neighbors >= n_samples:
        n_neighbors = n_samples - 1
    return n_neighbors


def _clone_modify(oversampler, y):
    """Clone and modify attributes of oversampler for corner cases."""

    # Clone oversampler
    oversampler = clone(oversampler)

    # Not modify attributes case
    if isinstance(oversampler, RandomOverSampler):
        return oversampler

    # Select and modify oversampler
    n_minority_samples = Counter(y).most_common()[-1][1]
    if n_minority_samples == 1:
        oversampler = RandomOverSampler()
    else:
        if hasattr(oversampler, "k_neighbors"):
            oversampler.k_neighbors = _modify_nn(
                oversampler.k_neighbors, n_minority_samples
            )
        if hasattr(oversampler, "m_neighbors"):
            oversampler.m_neighbors = _modify_nn(oversampler.m_neighbors, y.size)
        if hasattr(oversampler, "n_neighbors"):
            oversampler.n_neighbors = _modify_nn(
                oversampler.n_neighbors, n_minority_samples
            )
    return oversampler


class OverSamplingAugmentation(BaseOverSampler):
    """
    A wrapper to facilitate the use of `imblearn.over_sampling` objects for
    data augmentation.

    Parameters
    ----------
    oversampler : oversampler estimator, default=None
        Over-sampler to be used for data augmentation.

    augmentation_strategy : float, dict or {'oversampling', 'constant', 'proportional'}\
        , default='oversampling'
        Specifies how the data augmentation is done.

        - When ``float`` or ``int``, each class' frequency is augmented
          according to the specified ratio (which is equivalent to the ``proportional``
          strategy).

        - When ``oversampling``, the data augmentation is done according to the
          sampling strategy passed in the ``oversampler`` object. If ``value`` is not
          `None`, then the number of samples generated for each class equals the number
          of samples in the majority class multiplied by ``value``.

        - When ``constant``, each class frequency is augmented to match
          the value passed in the parameter ``value``.

        - When ``proportional``, relative class frequencies are preserved and the
          number of samples in the dataset is matched with the value passed in the
          parameter ``value``.

    value : int, float, default=None
        Value to be used as the new frequency of each class. It is ignored unless the
        augmentation strategy is set to ``constant`` or ``oversampling``.

    random_state : int, RandomState instance, default=None
        Control the randomization of the algorithm.

        - If int, ``random_state`` is the seed used by the random number
          generator;
        - If ``RandomState`` instance, random_state is the random number
          generator;
        - If ``None``, the random number generator is the ``RandomState``
          instance used by ``np.random``.

    Attributes
    ----------
    sampling_strategy_ : dict
        Dictionary containing the information to sample the dataset. The keys
        corresponds to the class labels from which to sample and the values
        are the number of samples to sample.
    n_features_in_ : int
        Number of features in the input dataset.
    """

    def __init__(
        self,
        oversampler=None,
        augmentation_strategy="oversampling",
        value=None,
        random_state=None,
    ):
        super(OverSamplingAugmentation, self).__init__(sampling_strategy="auto")
        self.oversampler = oversampler
        self.augmentation_strategy = augmentation_strategy
        self.value = value
        self.random_state = random_state

        warnings.filterwarnings("ignore")

    def fit(self, X, y):
        """
        Check inputs and statistics of the sampler.

        You should use ``fit_resample`` in all cases.

        Parameters
        ----------
        X : {array-like, dataframe, sparse matrix} of shape \
            (n_samples, n_features)
            Data array.

        y : array-like of shape (n_samples,)
            Target array.

        Returns
        -------
        self : object
            Return the instance itself.
        """

        X, y, _ = self._check_X_y(X, y)

        if (
            type(self.augmentation_strategy) not in [int, float, dict]
            and self.augmentation_strategy not in AUGMENTATION_STRATEGIES
        ):
            raise ValueError(
                f"When 'augmentation_strategy' is neither an int or float,"
                f" it needs to be one of {AUGMENTATION_STRATEGIES}. Got "
                f"'{self.augmentation_strategy}' instead."
            )

        if (type(self.value) not in [int, float]) and (
            self.augmentation_strategy in ["constant", "proportional"]
        ):
            raise ValueError(
                f"When 'augmentation_strategy' is 'constant' or 'proportional',"
                f" 'value' needs to be an int or float. Got "
                f"{self.value} instead."
            )

        # Setup the sampling strategy based on the augmentation strategy
        if self.augmentation_strategy == "constant":
            counts = OrderedDict(Counter(y))
            self.sampling_strategy_ = {
                k: int(np.round(self.value)) if self.value > freq else freq
                for k, freq in counts.items()
            }
        elif self.augmentation_strategy == "proportional":
            counts = OrderedDict(Counter(y))
            ratio = self.value / y.shape[0]
            if ratio > 1:
                self.sampling_strategy_ = {
                    k: int(np.round(freq * ratio)) for k, freq in counts.items()
                }
            else:
                raise ValueError(
                    "The new size of the augmented dataset must be larger than the"
                    f" original dataset. Originally, there are {y.shape[0]} samples"
                    f" and {self.value} samples are asked."
                )
        elif self.augmentation_strategy == "oversampling" and self.value is None:
            self.sampling_strategy_ = (
                self.oversampler.sampling_strategy
                if self.oversampler is not None
                else None
            )

        elif self.augmentation_strategy == "oversampling":
            counts = OrderedDict(Counter(y))
            max_freq = max(counts.values())
            self.sampling_strategy_ = {
                k: int(np.round(max_freq * self.value))
                if max_freq * self.value > freq
                else freq
                for k, freq in counts.items()
            }

        elif type(self.augmentation_strategy) in [int, float]:
            counts = OrderedDict(Counter(y))
            self.sampling_strategy_ = {
                k: int(np.round(v * self.augmentation_strategy))
                for k, v in counts.items()
            }
        else:
            self.sampling_strategy_ = self.augmentation_strategy

        return self

    def fit_resample(self, X, y, **fit_params):
        """
        Resample the dataset.

        Parameters
        ----------
        X : {array-like, dataframe, sparse matrix} of shape \
            (n_samples, n_features)
            Matrix containing the data which have to be sampled.

        y : array-like of shape (n_samples,)
            Corresponding label for each sample in X.

        Returns
        -------
        X_resampled : {array-like, dataframe, sparse matrix} of shape \
            (n_samples_new, n_features)
            The array containing the resampled data.

        y_resampled : array-like of shape (n_samples_new,)
            The corresponding label of `X_resampled`.
        """

        self.fit(X, y)

        if self.oversampler is not None:
            self.oversampler_ = _clone_modify(self.oversampler, y).set_params(
                random_state=self.random_state,
                sampling_strategy=self.sampling_strategy_,
            )
            if isinstance(self.oversampler_, GeometricSMOTE):
                return self.oversampler_.fit_resample(X, y, **fit_params)
            else:
                return self.oversampler_.fit_resample(X, y)
        else:
            return X, y

    def _fit_resample(self, X, y):
        """A placeholder. It was overriden by the self.fit_resample method."""
        return



import numpy as np
import matplotlib.pyplot as plt

from .._visualization import set_matplotlib_style, feature_to_color


def test_set_matplotlib_style():
    default_params = dict(plt.rcParams)
    set_matplotlib_style(12)
    new_params = dict(plt.rcParams)
    changed_params = [
        key for key in new_params.keys() if new_params[key] != default_params[key]
    ]
    assert len(changed_params) > 1


def test_feature_to_color():
    colors = feature_to_color(np.array([1, 2, 3, 4, 5]))
    colors2 = feature_to_color([1, 2, 3, 4, 5])
    assert (colors == colors2).all()
    assert colors.size == np.unique(colors).size



"""
Download, transform and simulate various datasets.
"""

# Author: Georgios Douzas <gdouzas@icloud.com>
#         Joao Fonseca <jpfonseca@novaims.unl.pt>
# License: MIT

from typing import Optional, Union
import os
from os.path import expanduser, join
from collections import Counter
from urllib.parse import urljoin
from sqlite3 import connect
from rich.progress import track
import numpy as np
import pandas as pd
from sklearn.utils import check_X_y
from imblearn.datasets import make_imbalance

UCI_URL = "https://archive.ics.uci.edu/ml/machine-learning-databases/"
GIC_URL = "http://www.ehu.eus/ccwintco/uploads/"
OPENML_URL = "https://www.openml.org/data/get_csv/"
FETCH_URLS = {
    "breast_tissue": urljoin(UCI_URL, "00192/BreastTissue.xls"),
    "ecoli": urljoin(UCI_URL, "ecoli/ecoli.data"),
    "eucalyptus": urljoin(OPENML_URL, "3625/dataset_194_eucalyptus.arff"),
    "glass": urljoin(UCI_URL, "glass/glass.data"),
    "haberman": urljoin(UCI_URL, "haberman/haberman.data"),
    "heart": urljoin(UCI_URL, "statlog/heart/heart.dat"),
    "iris": urljoin(UCI_URL, "iris/bezdekIris.data"),
    "libras": urljoin(UCI_URL, "libras/movement_libras.data"),
    "liver": urljoin(UCI_URL, "liver-disorders/bupa.data"),
    "pima": "https://gist.githubusercontent.com/ktisha/c21e73a1bd1700294ef790c56c8aec1f"
    "/raw/819b69b5736821ccee93d05b51de0510bea00294/pima-indians-diabetes.csv",
    "vehicle": urljoin(UCI_URL, "statlog/vehicle/"),
    "wine": urljoin(UCI_URL, "wine/wine.data"),
    "new_thyroid": urljoin(UCI_URL, "thyroid-disease/new-thyroid.data"),
    "cleveland": urljoin(UCI_URL, "heart-disease/processed.cleveland.data"),
    "led": urljoin(OPENML_URL, "4535757/phpSj3fWL"),
    "page_blocks": urljoin(OPENML_URL, "30/dataset_30_page-blocks.arff"),
    "yeast": urljoin(UCI_URL, "yeast/yeast.data"),
    "banknote_authentication": urljoin(
        UCI_URL, "00267/data_banknote_authentication.txt"
    ),
    "arcene": urljoin(UCI_URL, "arcene/"),
    "audit": urljoin(UCI_URL, "00475/audit_data.zip"),
    "spambase": urljoin(UCI_URL, "spambase/spambase.data"),
    "parkinsons": urljoin(UCI_URL, "parkinsons/parkinsons.data"),
    "ionosphere": urljoin(UCI_URL, "ionosphere/ionosphere.data"),
    "breast_cancer": urljoin(UCI_URL, "breast-cancer-wisconsin/wdbc.data"),
    "adult": urljoin(UCI_URL, "adult/adult.data"),
    "abalone": urljoin(UCI_URL, "abalone/abalone.data"),
    "acute": urljoin(UCI_URL, "acute/diagnosis.data"),
    "annealing": urljoin(UCI_URL, "annealing/anneal.data"),
    "census": urljoin(UCI_URL, "census-income-mld/census-income.data.gz"),
    "contraceptive": urljoin(UCI_URL, "cmc/cmc.data"),
    "covertype": urljoin(UCI_URL, "covtype/covtype.data.gz"),
    "credit_approval": urljoin(UCI_URL, "credit-screening/crx.data"),
    "dermatology": urljoin(UCI_URL, "dermatology/dermatology.data"),
    "echocardiogram": urljoin(UCI_URL, "echocardiogram/echocardiogram.data"),
    "flags": urljoin(UCI_URL, "flags/flag.data"),
    "heart_disease": [
        urljoin(UCI_URL, "heart-disease/processed.cleveland.data"),
        urljoin(UCI_URL, "heart-disease/processed.hungarian.data"),
        urljoin(UCI_URL, "heart-disease/processed.switzerland.data"),
        urljoin(UCI_URL, "heart-disease/processed.va.data"),
    ],
    "hepatitis": urljoin(UCI_URL, "hepatitis/hepatitis.data"),
    "german_credit": urljoin(UCI_URL, "statlog/german/german.data"),
    "thyroid": urljoin(UCI_URL, "thyroid-disease/thyroid0387.data"),
    "first_order_theorem": urljoin(OPENML_URL, "1587932/phpPbCMyg"),
    "gas_drift": urljoin(OPENML_URL, "1588715/phpbL6t4U"),
    "autouniv_au7": urljoin(OPENML_URL, "1593748/phpmRPvKy"),
    "autouniv_au4": urljoin(OPENML_URL, "1593744/phpiubDlf"),
    "mice_protein": urljoin(OPENML_URL, "17928620/phpchCuL5"),
    "steel_plates": urljoin(OPENML_URL, "18151921/php5s7Ep8"),
    "cardiotocography": urljoin(OPENML_URL, "1593756/phpW0AXSQ"),
    "waveform": urljoin(OPENML_URL, "60/dataset_60_waveform-5000.arff"),
    "volkert": urljoin(OPENML_URL, "19335689/file1c556e3db171.arff"),
    "asp_potassco": urljoin(OPENML_URL, "21377447/file18547f421393.arff"),
    "wine_quality": urljoin(OPENML_URL, "4965268/wine-quality-red.arff"),
    "mfeat_zernike": urljoin(OPENML_URL, "22/dataset_22_mfeat-zernike.arff"),
    "gesture_segmentation": urljoin(OPENML_URL, "1798765/phpYLeydd"),
    "texture": urljoin(OPENML_URL, "4535764/phpBDgUyY"),
    "usps": urljoin(OPENML_URL, "19329737/usps.arff"),
    "vowels": urljoin(OPENML_URL, "52415/JapaneseVowels.arff"),
    "pendigits": urljoin(OPENML_URL, "32/dataset_32_pendigits.arff"),
    "image_segmentation": urljoin(OPENML_URL, "18151937/phpyM5ND4"),
    "baseball": urljoin(OPENML_URL, "3622/dataset_189_baseball.arff"),
    "indian_pines": [
        urljoin(GIC_URL, "2/22/Indian_pines.mat"),
        urljoin(GIC_URL, "c/c4/Indian_pines_gt.mat"),
    ],
    "salinas": [
        urljoin(GIC_URL, "f/f1/Salinas.mat"),
        urljoin(GIC_URL, "f/fa/Salinas_gt.mat"),
    ],
    "salinas_a": [
        urljoin(GIC_URL, "d/df/SalinasA.mat"),
        urljoin(GIC_URL, "a/aa/SalinasA_gt.mat"),
    ],
    "pavia_centre": [
        urljoin(GIC_URL, "e/e3/Pavia.mat"),
        urljoin(GIC_URL, "5/53/Pavia_gt.mat"),
    ],
    "pavia_university": [
        urljoin(GIC_URL, "e/ee/PaviaU.mat"),
        urljoin(GIC_URL, "5/50/PaviaU_gt.mat"),
    ],
    "kennedy_space_center": [
        urljoin(GIC_URL, "2/26/KSC.mat"),
        urljoin(GIC_URL, "a/a6/KSC_gt.mat"),
    ],
    "botswana": [
        urljoin(GIC_URL, "7/72/Botswana.mat"),
        urljoin(GIC_URL, "5/58/Botswana_gt.mat"),
    ],
}
RANDOM_STATE = 0


def get_data_home(data_home: Optional[str] = None) -> str:
    """Return the path of the ml-research data dir.
    This folder is used by some large dataset loaders to avoid downloading the
    data several times.
    By default the data dir is set to a folder named 'ml_research_data' in the
    user home folder.
    Alternatively, it can be set programmatically by giving an explicit folder
    path. The '~' symbol is expanded to the user home folder.
    If the folder does not already exist, it is automatically created.

    Parameters
    ----------
    data_home : str, default=None
        The path to the data directory. If `None`, the default path
        is `~/ml_research_data`.
    """
    if data_home is None:
        data_home = join("~", "mlresearch_data")
    data_home = expanduser(data_home)
    os.makedirs(data_home, exist_ok=True)
    return data_home


class Datasets:
    """
    Base class to download and save datasets.
    """

    def __init__(
        self,
        names: Union[str, list] = "all",
        data_home: str = None,
        download_if_missing: bool = True,
    ):
        """
        Parameters
        ----------
        names : str or list, default="all"
            List of dataset names to be downloaded. If ``all``, downloads all datasets.

        data_home : str, default=None
            The path to the data directory. If `None`, the default path
            is `~/ml_research_data`.

        download_if_missing : bool, default=True
            If True, downloads the dataset from the internet and puts it in
            ``data_home``. If the dataset is already downloaded, it is not downloaded
            again.

        Attributes
        ----------
        data_home_ : str
            Path were the data was stored.

        content_ : list
            List of tuples composed of (Dataset name, Dataframe).
        """
        self.names = names
        self.data_home = data_home
        self.download_if_missing = download_if_missing

    @staticmethod
    def _modify_columns(data, keep_index=False):
        """Rename and reorder columns of dataframe."""
        X, y = data.drop(columns="target"), data.target
        X.columns = X.columns if keep_index else range(len(X.columns))
        return pd.concat([X, y], axis=1)

    @staticmethod
    def _calculate_sampling_strategy(ir, y):
        """Calculate ratio based on the IR."""
        ratio = Counter(y).most_common()
        b = ratio[0][1]
        c_min = ratio[0][1] / ir
        m = (c_min - b) / (len(ratio) - 1)

        sampling_strategy = {}
        for x, (c, f) in enumerate(ratio):
            f_new = int(np.clip(m * x + b, 1, f))
            sampling_strategy[c] = f_new

        return sampling_strategy

    def _make_imbalance(self, data, sampling_strategy, random_state=None):
        """Undersample the minority class."""
        X, y = check_X_y(data.drop(columns="target"), data["target"], dtype=None)
        X, y = make_imbalance(
            X, y, sampling_strategy=sampling_strategy, random_state=random_state
        )
        data = pd.DataFrame(np.column_stack((X, y)))
        data[data.columns[-1]] = data[data.columns[-1]].astype(int)
        return data

    def download(self, keep_index=False):
        """Download the datasets."""
        self.data_home_ = get_data_home(data_home=self.data_home)
        dataset_prefix = self.__class__.__name__.lower().replace("datasets", "")

        # Get datasets to download
        if self.names == "all":
            func_names = [func_name for func_name in dir(self) if "fetch_" in func_name]
        else:
            func_names = [
                f"fetch_{name}".lower().replace(" ", "_") for name in self.names
            ]

        # Download datasets
        self.content_ = []
        for func_name in track(func_names, description="Datasets"):
            dat_name = func_name.replace("fetch_", "")
            name = dat_name.upper().replace("_", " ")
            file_name = f"{dataset_prefix}_{dat_name}.csv"

            # Save file locally if it doesn't exist
            if (
                file_name not in os.listdir(self.data_home_)
                and self.download_if_missing
            ):
                df = getattr(self, func_name)()
                df.to_csv(join(self.data_home_, file_name), index=keep_index)

            data = pd.read_csv(join(self.data_home_, file_name))
            data = self._modify_columns(data, keep_index=keep_index)
            self.content_.append((name, data))
        return self

    def imbalance_datasets(self, imbalance_ratio: float, random_state: int = None):
        """
        Appends imbalanced versions of datasets with predefined imbalance ratios to
        ``self.content_``.

        .. math::
            IR = \\frac{|C_{maj}|}{|C_{min}|}

        Parameters
        ----------
        imbalance_ratio : float
            Final Imbalance Ratio expected in the datasets.

        random_state : int, RandomState instance, default=None
            Control the randomization of the algorithm.

            - If int, ``random_state`` is the seed used by the random number
              generator;
            - If ``RandomState`` instance, random_state is the random number
              generator;
            - If ``None``, the random number generator is the ``RandomState``
              instance used by ``np.random``.

        Returns
        -------
        self : Datasets

        """
        imbalanced_content = []
        base_content = [
            dataset for dataset in self.content_ if not dataset[0].endswith(")")
        ]
        for name, data in base_content:
            base_freqs = Counter(data.target).values()
            base_ir = int(max(base_freqs) / min(base_freqs))
            sampling_strategy = self._calculate_sampling_strategy(
                ir=imbalance_ratio, y=data.target
            )
            data_imb = self._make_imbalance(
                data, sampling_strategy=sampling_strategy, random_state=random_state
            )

            data_imb.columns = data.columns

            freqs = Counter(data_imb.target).values()
            new_ir = int(max(freqs) / min(freqs))
            name_imb = f"{name} ({new_ir})"
            if name_imb not in dict(self.content_).keys() and base_ir < new_ir:
                imbalanced_content.append((name_imb, data_imb))

        self.content_.extend(imbalanced_content)
        return self

    def summarize_datasets(self):
        """
        Create a summary of the downloaded datasets.

        Returns
        -------
        datasets_summary : pd.DataFrame
            Dataframe with summary statistics of all datasets.
        """

        # Check datasets format
        datasets = [
            (name, (data.drop(columns="target"), data.target))
            if type(data) == pd.DataFrame
            else data
            for name, data in self.content_
        ]

        # Define summary table columns
        summary_columns = [
            "Dataset name",
            "Features",
            "Observations",
            "Minority Obs.",
            "Majority Obs.",
            "Imbalance Ratio",
            "Classes",
        ]

        # Define empty summary table
        datasets_summary = []

        # Populate summary table
        for dataset_name, (X, y) in datasets:
            n_instances = Counter(y).values()
            n_minority_instances = min(n_instances)
            n_majority_instances = max(n_instances)
            values = [
                dataset_name,
                X.shape[1],
                len(X),
                n_minority_instances,
                n_majority_instances,
                round(n_majority_instances / n_minority_instances, 2),
                len(n_instances),
            ]
            datasets_summary.append(values)
        datasets_summary = pd.DataFrame(datasets_summary, columns=summary_columns)

        # Cast to integer columns
        int_cols = datasets_summary.columns.drop(["Dataset name", "Imbalance Ratio"])
        datasets_summary.loc[:, int_cols] = datasets_summary.loc[:, int_cols].astype(
            int
        )

        # Sort datasets summary
        datasets_summary = datasets_summary.sort_values("Imbalance Ratio").reset_index(
            drop=True
        )

        return datasets_summary

    def save(self, path, db_name):
        """Save datasets."""
        with connect(join(path, f"{db_name}.db")) as connection:
            for name, data in self.content_:
                data.to_sql(name, connection, index=False, if_exists="replace")



"""
Module which contains the implementation of variations of oversampling/data augmentation
algorithms, as well as helper classes to use oversampling algorithms as data augmentation
techniques.
"""

from ._oversampling_augmentation import OverSamplingAugmentation
from ._gsmote import GeometricSMOTE

__all__ = ["OverSamplingAugmentation", "GeometricSMOTE"]



from sklearn.base import ClassifierMixin
from .base import BaseActiveLearner


class SelfSupervisedAL(BaseActiveLearner, ClassifierMixin):
    """
    Active Learning with self supervision as presented in [1]_. The model is pretrained
    using SimSiam [2]_. They use the pretrained model to extract features, which are used
    to train a SVM or a Logistic Regression.

    Notes
    -----
    See the original paper: [1]_ for more details.

    References
    ----------
    .. [1] Bengar, J. Z., van de Weijer, J., Twardowski, B., & Raducanu, B. (2021).
       Reducing label effort: Self-supervised meets active learning. In Proceedings of
       the IEEE/CVF International Conference on Computer Vision (pp. 1631-1639).

    .. [2] Chen, X., & He, K. (2021). Exploring simple siamese representation learning.
       In Proceedings of the IEEE/CVF Conference on Computer Vision and Pattern
       Recognition (pp. 15750-15758).
    """


class LADA(BaseActiveLearner, ClassifierMixin):
    """Look-Ahead Data Augmentation for Active Learning implementation."""


class LearningLoss(BaseActiveLearner, ClassifierMixin):
    """Learning Loss implementation."""


class CoreSet(BaseActiveLearner, ClassifierMixin):
    """Coreset Active Learning implementation."""
